#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>

import os
import sys
import gi
gi.require_version("Gtk", "3.0")  # @UndefinedVariable
gi.require_version("Gdk", "3.0")  # @UndefinedVariable
from gi.repository import Gtk, Gdk, GdkPixbuf  # pylint: disable=wrong-import-position @UnresolvedImport

count = 0
def handle_owner_change(clipboard, event):
    global count
    print('clipboard.owner-change(%r, %r)' % (clipboard, event))
    #count += 1
    #if count > 1:
    #    sys.exit(0)

def main():
    assert os.path.exists(sys.argv[1])
    image = GdkPixbuf.Pixbuf.new_from_file(sys.argv[1])
    clipboard = Gtk.Clipboard.get(Gdk.Atom.intern("CLIPBOARD", False))
    clipboard.connect('owner-change', handle_owner_change)
    clipboard.set_image(image)
    clipboard.store()
    Gtk.main()

if __name__ == "__main__":
    main()
    sys.exit(0)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=line-too-long

import unittest

from xpra.os_util import WIN32, POSIX, OSX
from xpra.util.objects import AdHocStruct
from xpra.client.mixins.audio import AudioClient
from xpra.audio.gstreamer_util import CODEC_ORDER
from unit.client.mixins.clientmixintest_util import ClientMixinTest


def default_audio_options() -> AdHocStruct:
    opts = AdHocStruct()
    opts.av_sync = True
    opts.speaker = "no"
    opts.microphone = "no"
    opts.audio_source = ""
    opts.speaker_codec = []
    opts.microphone_codec = []
    opts.tray_icon = ""
    return opts


class AudioClientTestUtil(ClientMixinTest):

    @classmethod
    def setUpClass(cls):
        ClientMixinTest.setUpClass()
        from xpra.net import packet_encoding
        packet_encoding.init_all()
        from xpra.net import compression
        compression.init_all()

    def _test_audio(self, opts, caps):
        return self._test_mixin_class(AudioClient, opts, caps)


class AudioClientSendTestUtil(AudioClientTestUtil):

    def do_test_audio_send(self, auto_start=True):
        opts = default_audio_options()
        opts.microphone = "on" if auto_start else "off"
        self._test_audio(opts, {
            "audio": {
                "receive": True,
                "decoders": CODEC_ORDER,
            },
        })
        if not self.mixin.microphone_codecs:
            print("no microphone codecs, test skipped")
            return

        def check_packets():
            if len(self.packets) < 5:
                return True
            self.mixin.stop_sending_audio()
            self.main_loop.quit()
            return False
        if not auto_start:
            def request_start():
                self.mixin.start_sending_audio()
            self.glib.timeout_add(500, request_start)
        self.glib.timeout_add(100, check_packets)
        self.glib.timeout_add(5000, self.main_loop.quit)
        self.main_loop.run()
        assert len(self.packets) > 2
        self.verify_packet(0, ("sound-data", ))
        assert self.packets[0][3].get("start-of-stream"), "start-of-stream not found"
        self.verify_packet(-1, ("sound-data", ))
        assert self.packets[-1][3].get("end-of-stream"), "end-of-stream not found"


class AudioClientSendAuto(AudioClientSendTestUtil):

    def test_audio_send_auto(self):
        self.do_test_audio_send(True)


class AudioClientSendRequest(AudioClientSendTestUtil):

    def test_audio_send_request(self):
        self.do_test_audio_send(False)


class AudioClientReceiveTest(AudioClientTestUtil):

    def test_audio_receive(self):
        opts = default_audio_options()
        opts.speaker = "yes"
        x = self._test_audio(opts, {
            "audio": {
                "send": True,
                "encoders": CODEC_ORDER,
            },
        })

        def stop() -> None:
            x.stop_receiving_audio()
            self.stop()
        if not self.mixin.speaker_codecs:
            stop()
            print("no speaker codecs, test skipped")
        if "opus" not in self.mixin.speaker_codecs:
            stop()
            print("'opus' speaker codec missing, test skipped")
            return

        packet_data = [
            ('sound-data', 'opus', '', {'start-of-stream': True, 'codec': b'opus'}),
            ('sound-data', 'opus', b'fc60fddad634b19a5baa6dd0ae26e05d15106df1135c84590fa2ab85d9945dd504a1d7b54d3ea189b276b36909ee33d34f038fd0d4aa25baa6abd1cd6b896bbfab52e02b9b18bc260fc4441bc44a65b2e0428431aafeabc3cc4974c9a4afe02df92638c51c1eb36292662b710ee971fb16361692e6fb819a1ef7bc66a6badd04d71160c16b249aaf497e79cf56c622cffeb6bcfd83027954132d5d90500104a4',
             {b'duration': 13500000, b'timestamp': 0, b'time': 159963539, 'sequence': 0}),
            ('sound-data', 'opus', b'fc6e8120f7b82efd4dec6f067d0f4af7ee27220307090f8595517139761252409ea93e4ba08d59b009a96efabc9a99f5895c2b084db0ab0ccc9665dc54e58a3a1dcacb6156fab84b6031c74b4f353ffb16c68c0959510e63d1f632a95358ee673d969c210f18b1dd765204a857c631e281960b9988d4821a34bf5d0729cf4696d8a00dfa08115cdd6de84b9bc8f216d6924e071018ebb6d4ac3c3d8298cb1813',
             {b'duration': 20000000, b'timestamp': 13500000, b'time': 159963558, 'sequence': 0}),
            ('sound-data', 'opus', b'fc6dd5d0f08acc6ae38aca6dc1038b2da49ca1fa69220405c8556f553e6352680ad0c403b94285a32259dd9455d157090ab602ac2256b4426779083ed9f119fe261de3fb7bafd3e1508cf38b47a7ff62d8d15443c095bd8ca63d152a6b1dcce71d70c79763163c2ed5e804eafffa0eb267b59bb35a1a34bdd7074a8f1696d8a406f0b4c457377ebc25cdadb789b5a49381c4063a44c7b6d4ac2bebebef316d77',
             {b'duration': 20000000, b'timestamp': 33500000, b'time': 159963576, 'sequence': 0}),
            ('sound-data', 'opus', b'fc4d2dfba53a9041a76a560e128ed23c27e6ba1a5dd7aaf9fa0ef2c801e748dc8a7262ca5821ccee59a209ccd97be931a62b02651d11af09e7d26aedbbb52a490a75463dccc52843b488e9ffd8b45ff6e5fe3cbcb4e872170cee673a82d953af2e3ce570ca82176af40df8db59735cdcdecfe0444d0faeebce583aa70241c05d1406f0c06d7bddd6defadc6b6de26d6924e071017dbfdb25645f0055f3d569a5',
             {b'duration': 20000000, b'timestamp': 53500000, b'time': 159963596, 'sequence': 0}),
            ('sound-data', 'opus', b'fc4d2e0df8779efc983d56a4881571c36a50ce90597b0be6a8d4d9194d5620c2c0bd0b9d7ecb0b8e724e7a9acbb9e9cf1d5ec31635f1bfa99ec410c49fc09bed96b9dd5a9b372c0e09a9ffd8ab8bfe3cbcdb97f2170cd3a1ddcce7505b2824c81aefefec32a085dabd037e36d65f5fd652f9c33d96673b2bcedf4d2a21d7d5dd1406f0c06d7bddd6defadc6b6de26d6924e071017d8fdb25645f50105c730413',
             {b'duration': 20000000, b'timestamp': 73500000, b'time': 159963616, 'sequence': 0}),
            ('sound-data', 'opus', '', {'end-of-stream': True, 'sequence': 0}),
        ]
        L = len(packet_data)

        def feed_data() -> bool:
            packet = packet_data.pop(0)
            self.handle_packet(packet)
            if packet_data:
                return True
            self.mixin.stop_receiving_audio()
            self.main_loop.quit()
            return False

        def check_start() -> bool:
            if not self.packets:
                return True
            self.verify_packet(0, ("sound-control", "start", "opus"))
            self.glib.timeout_add(100, feed_data)
            return False

        self.glib.timeout_add(100, check_start)
        self.glib.timeout_add(5000, stop)
        # self.debug_all()
        self.main_loop.run()
        assert len(packet_data) < L, "none of the data was fed to the receiver"
        assert not packet_data, "not all the data was fed to the receiver: remains: %s" % len(packet_data)
        self.verify_packet(0, ("sound-control", "start", "opus"))
        self.verify_packet(1, ("sound-control", "new-sequence", 1))
        # assert not self.packets, "sent some unexpected packets: %s" % (self.packets,)
        assert self.mixin.audio_sink is None, "sink is still active: %s" % self.mixin.audio_sink


def main():
    if WIN32:
        return
    if POSIX and not OSX:
        # verify that pulseaudio is running:
        # otherwise the tests will fail
        # ie: during rpmbuild
        from subprocess import getstatusoutput
        if getstatusoutput("pactl info")[0]!=0:
            return
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import hmac
import hashlib
from collections.abc import Callable, Sequence

from xpra.common import SizedBuffer
from xpra.util.str_fn import csv, strtobytes, hexstr, memoryview_to_bytes
from xpra.util.env import envint
from xpra.log import Logger

log = Logger("network", "crypto")

BLOCKLISTED_HASHES = ("sha1", "md5")
DEFAULT_SALT_LENGTH = envint("XPRA_DEFAULT_SALT_LENGTH", 64)


def get_digests() -> Sequence[str]:
    digests = ["xor"]
    digests += [f"hmac+{x}" for x in tuple(reversed(sorted(hashlib.algorithms_available)))
                if not x.startswith("shake_") and x not in BLOCKLISTED_HASHES and getattr(hashlib, x, None) is not None]
    try:
        from xpra.net.rfb import d3des  # pylint: disable=import-outside-toplevel
        assert d3des
        digests.append("des")
    except (ImportError, TypeError):  # pragma: no cover
        pass
    return tuple(digests)


def get_digest_module(digest: str) -> Callable | None:
    log(f"get_digest_module({digest})")
    if not digest or not digest.startswith("hmac"):
        return None
    try:
        digest_module = digest.split("+")[1]  # ie: "hmac+sha512" -> "sha512"
    except IndexError:
        return None
    try:
        return getattr(hashlib, digest_module)
    except AttributeError as e:
        log("no '%s' attribute in hashlib: %s", digest_module, e)
        return None


def choose_digest(options: Sequence[str]) -> str:
    assert len(options) > 0, "no digest options"
    log(f"choose_digest({options})")
    # prefer stronger hashes:
    for h in ("sha512", "sha384", "sha256", "sha224"):
        hname = f"hmac+{h}"
        if hname in options:
            return hname
    if "xor" in options:
        return "xor"
    if "des" in options:
        return "des"
    raise ValueError(f"no known digest options found in '{csv(options)}'")


def gendigest(digest: str, password_in, salt_in: SizedBuffer) -> bytes:
    assert password_in and salt_in
    salt: bytes = memoryview_to_bytes(salt_in)
    password: bytes = strtobytes(password_in)
    if digest == "des":
        from xpra.net.rfb.d3des import generate_response  # pylint: disable=import-outside-toplevel
        password = password.ljust(8, b"\x00")[:8]
        salt = salt.ljust(16, b"\x00")[:16]
        v = generate_response(password, salt)
        return strtobytes(hexstr(v))
    if digest in ("xor", "kerberos", "gss", "keycloak"):
        # kerberos, gss and keycloak use xor because we need to use the actual token
        # at the other end
        salt = salt.ljust(len(password), b"\x00")[:len(password)]
        from xpra.buffers.cyxor import xor_str  # pylint: disable=import-outside-toplevel
        v = xor_str(password, salt)
        return memoryview_to_bytes(v)
    digestmod = get_digest_module(digest)
    if not digestmod:
        log(f"invalid digest module {digest!r}")
        return b""
        # warn_server_and_exit(ExitCode.UNSUPPORTED,
        #    "server requested digest '%s' but it is not supported" % digest, "invalid digest")
    return strtobytes(hmac.new(password, salt, digestmod=digestmod).hexdigest())


def verify_digest(digest: str, password: str, salt, challenge_response: bytes) -> bool:
    log("verify_digest%s", (digest, "..", salt, challenge_response))
    if not (digest and password and salt and challenge_response):
        log("digest: missing argument")
        return False
    verify = gendigest(digest, password, salt)
    verified = hmac.compare_digest(verify, challenge_response)
    log("verify_digest(..) %r vs %r", verify, challenge_response)
    if not verified:
        log(f"expected {verify!r} but got {challenge_response!r}")
    return verified


def get_salt(length: int = DEFAULT_SALT_LENGTH) -> bytes:
    # too short: we would not feed enough random data to HMAC
    if length < 32:
        raise ValueError(f"salt is too short: only {length} bytes")
    # too long: limit the amount of random data we request from the system
    if length >= 1024:
        raise ValueError(f"salt is too long: {length} bytes")
    # all server versions support a client salt,
    # they also tell us which digest to use:
    return os.urandom(length)



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from importlib import import_module


_init_done = False


def init(prgname="", appname="") -> None:
    """ do whatever is needed to prepare an application for running,
        some platforms may initialize logging to file, etc.
        If the names are supplied, we call set_name()
    """
    global _init_done
    if prgname or appname:
        set_default_name(prgname, appname)
        set_name()
    init_env()
    if not _init_done:
        _init_done = True
        do_init()


def do_init() -> None:  # pragma: no cover
    """ some platforms override this """


def init_env() -> None:
    do_init_env()


def do_init_env() -> None:
    init_env_common()


def init_env_common() -> None:
    # turn off gdk scaling to make sure we get the actual window geometry:
    os.environ["GDK_SCALE"] = os.environ.get("GDK_SCALE", "1")
    os.environ["GDK_DPI_SCALE"] = os.environ.get("GDK_DPI_SCALE", "1")
    # client side decorations break window geometry,
    # disable this "feature" unless explicitly enabled:
    os.environ["GTK_CSD"] = os.environ.get("GTK_CSD", "0")
    init_hashlib()


def init_hashlib() -> None:
    from xpra.util.env import envbool
    if envbool("XPRA_NOMD5", False):
        import hashlib
        try:
            hashlib.algorithms_available.remove("md5")  # type: ignore[attr-defined] #@UndefinedVariable
        except KeyError:
            pass
        else:
            def nomd5(*_anyargs):
                raise ValueError("md5 support is disabled")
            hashlib.md5 = nomd5                         # type: ignore
    if envbool("XPRA_NOSHA1", False):
        import hashlib  # @Reimport
        try:
            hashlib.algorithms_available.remove("sha1")  # type: ignore[attr-defined] #@UndefinedVariable
        except KeyError:
            pass
        else:
            def nosha1(*_anyargs):
                raise ValueError("sha1 support is disabled")
            hashlib.sha1 = nosha1                       # type: ignore


def threaded_server_init():
    """ platform implementations may override this function """


class program_context:
    def __init__(self, prgname="", appname=""):
        self.prgname = prgname
        self.appname = appname

    def __enter__(self):
        init(self.prgname, self.appname)
        return self

    def __exit__(self, *_args):
        clean()

    def __repr__(self):
        return f"gui_context({self.prgname}, {self.appname})"


_prgname = ""
_appname = ""


def set_default_name(prgname="", appname="") -> None:
    # sets the default prg and app names
    global _prgname, _appname
    if prgname:
        _prgname = prgname
    if appname:
        _appname = appname


# platforms can override this
def command_error(message) -> None:
    from xpra.scripts.main import error
    error(message)


def command_info(message) -> None:
    from xpra.scripts.main import info
    info(message)


_clean_done = False


def clean() -> None:
    global _clean_done
    if not _clean_done:
        _clean_done = True
        do_clean()


def do_clean():  # pragma: no cover
    """ some platforms override this """


_name_set = False


def set_name(prgname="", appname="") -> None:
    global _name_set
    if not _name_set:
        _name_set = True
        set_prgname(prgname or _prgname)
        set_application_name(appname or _appname)


# platforms can override this
def set_prgname(name="") -> None:
    if not name:
        return
    try:
        from xpra.os_util import gi_import
        GLib = gi_import("GLib")
        GLib.set_prgname(name)
    except ImportError:
        pass


def get_prgname() -> str:
    global _prgname
    return _prgname


# platforms can override this
def set_application_name(name="") -> None:
    if not name:
        return
    try:
        from xpra.os_util import gi_import
        GLib = gi_import("GLib")
    except ImportError:
        pass
    else:
        GLib.set_application_name(name)


def get_application_name() -> str:
    global _appname
    return _appname


def is_terminal() -> bool:
    return bool(os.environ.get("TERM", ""))


def platform_import(where: dict, pm="", required=False, *imports) -> None:
    from xpra.os_util import OSX, POSIX
    if os.name == "nt":     # pragma: no cover
        p = "win32"
    elif OSX:               # pragma: no cover
        p = "darwin"
    elif POSIX:             # pragma: no cover
        p = "posix"
    else:                   # pragma: no cover
        raise OSError(f"Unknown OS {os.name!r}")

    module = "xpra.platform.%s" % p
    if pm:
        module += ".%s" % pm

    # cannot log this early! (win32 needs log to file redirection…)
    # log = Logger("platform", "import")
    # log("importing %s from %s (required=%s)" % (imports, module, required))
    try:
        platform_module = import_module(module)
    except ImportError:
        if required:
            raise
        return
    for x in imports:
        found = hasattr(platform_module, x)
        if not found:
            if required:
                raise ImportError(f"could not find {x} in {module}")
        else:
            where[x] = getattr(platform_module, x)


platform_import(globals(), "", False,
                "do_init", "do_clean", "do_init_env",
                "threaded_server_init",
                "set_prgname", "set_application_name", "program_context",
                "command_error", "command_info",
                "is_terminal",
                )



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import uuid
import signal
from typing import Any
from time import monotonic
from collections.abc import Callable

from pyglet import app, clock

from xpra import __version__
from xpra.common import noop
from xpra.exit_codes import ExitValue
from xpra.net.protocol.factory import get_client_protocol_class
from xpra.net.common import PacketType
from xpra.log import Logger

log = Logger("client")
netlog = Logger("client", "network")


class scheduled:
    def __init__(self, fn: Callable, args, kwargs):
        self.function = fn
        self.args = args
        self.kwargs = kwargs

    def run(self, elapsed: float):
        try:
            log(f"scheduled.run({elapsed}) calling {self.function}{self.args}{self.kwargs}")
            ret = self.function(*self.args, **self.kwargs)
        except Exception:
            log(f"scheduled error calling {self.function}{self.args}{self.kwargs}", exc_info=True)
            ret = False
        if not ret:
            clock.unschedule(self.run)


class PygletScheduler:

    def __init__(self):
        self.scheduled = []

    def idle_add(self, fn: Callable, *args, **kwargs):
        return self.timeout_add(0, fn, *args, **kwargs)

    def timeout_add(self, delay: int, fn: Callable, *args, **kwargs):
        # interval is in seconds:
        interval = delay / 1000
        instance = scheduled(fn, args, kwargs)
        self.scheduled.append(instance)
        log(f"pyglet_clock_scheduler: {instance} in {interval} seconds")
        clock.schedule_interval(instance.run, interval)


class XpraPygletClient:

    def __init__(self):
        self.windows: dict[int, Any] = {}
        self.hello = {}
        self.focused = 0
        self._ordinary_packets = []
        self.protocol = None
        self.have_more = noop

    def show_progress(self, pct: int, msg) -> None:
        log(f"show_progress({pct}, {msg})")

    def run(self) -> int:
        app.run()
        return 0

    def quit(self, exit_code: ExitValue):
        sys.exit(int(exit_code))

    def init(self, opts) -> None:
        """ we don't handle any options yet! """

    def init_ui(self, opts) -> None:
        """ we don't handle any options yet! """

    def cleanup(self) -> None:
        """ client classes must define this method """

    def make_protocol(self, conn):
        protocol_class = get_client_protocol_class(conn.socktype)
        protocol = protocol_class(conn, self.process_packet, self.next_packet, scheduler=PygletScheduler())
        log(f"setup_connection({conn}) {protocol=}")
        protocol.enable_default_encoder()
        protocol.enable_default_compressor()
        self._protocol = protocol
        self.have_more = protocol.source_has_more
        clock.schedule_once(self.send_hello, 1)
        return protocol

    def send_hello(self, elapsed) -> None:
        log(f"send_hello({elapsed})")
        hello = {
            "version": __version__,
            "client_type": "pyglet",
            "rencodeplus": True,
            "session-id": uuid.uuid4().hex,
            "windows": True,
            "keyboard": True,
            "mouse": True,
            "encodings": ("png", "jpg", "webp"),    # "rgb32", "rgb24"
            "network-state": False,  # tell older server that we don't have "ping"
        }
        self.send("hello", hello)

    def send(self, packet_type: str, *args) -> None:
        packet = (packet_type, *args)
        # data = dumps(packet)
        # header = pack_header(FLAGS_RENCODEPLUS, 0, 0, len(data))
        # bin_packet = header+data
        netlog(f"sent {packet_type!r}: {packet!r}")
        self._ordinary_packets.append(packet)
        self.have_more()

    def next_packet(self) -> tuple[PacketType, bool, bool]:
        packet = self._ordinary_packets.pop(0)
        return packet, True, bool(self._ordinary_packets)

    def process_packet(self, _protocol, packet: PacketType) -> None:
        packet_type_fn_name = str(packet[0]).replace("-", "_")
        meth = getattr(self, f"_process_{packet_type_fn_name}", None)
        if not meth:
            netlog.warn(f"Warning: missing handler for {packet[0]!r}")
            netlog("packet=%r", packet)
            return

        def call_handler(_elapsed) -> None:
            meth(packet)
        clock.schedule_once(call_handler, 0)

    def _process_hello(self, packet: PacketType) -> None:
        self.hello = packet[1]
        netlog.info("got hello from %s server" % self.hello.get("version", ""))

    def _process_setting_change(self, packet: PacketType) -> None:
        setting = packet[1]
        netlog.info(f"ignoring setting-change for {setting!r}")

    def _process_startup_complete(self, _packet: PacketType) -> None:
        netlog.info("client is connected")

    def _process_encodings(self, packet: PacketType) -> None:
        log(f"server encodings: {packet[1]}")

    def _process_new_window(self, packet: PacketType) -> None:
        self.new_window(packet)

    def _process_new_override_redirect(self, packet: PacketType) -> None:
        self.new_window(packet, True)

    def new_window(self, packet: PacketType, is_or=False) -> None:
        from xpra.client.pyglet.window import ClientWindow
        wid, x, y, w, h = (int(item) for item in packet[1:6])
        metadata = packet[6]
        if is_or:
            metadata["override-redirect"] = is_or
        window = ClientWindow(self, wid, x, y, w, h, metadata)
        self.windows[wid] = window
        window.show()

    def _process_lost_window(self, packet: PacketType) -> None:
        wid = int(packet[1])
        window = self.windows.get(wid)
        if window:
            window.close()
            del self.windows[wid]

    def _process_raise_window(self, packet: PacketType) -> None:
        wid = int(packet[1])
        window = self.windows.get(wid)
        if window:
            window.raise_()

    def _process_draw(self, packet: PacketType) -> None:
        wid, x, y, width, height, coding, data, packet_sequence, rowstride = packet[1:10]
        window = self.windows.get(wid)
        now = monotonic()
        if window:
            message = ""
            window.draw(x, y, width, height, coding, data, rowstride)
            decode_time = int(1000 * (monotonic() - now))
        else:
            message = f"Warning: window {wid} not found"
            log.warn(message)
            decode_time = -1
        self.send("damage-sequence", packet_sequence, wid, width, height, decode_time, message)

    def _process_window_metadata(self, packet: PacketType) -> None:
        wid = packet[1]
        metadata = packet[2]
        log.info(f"window {wid}: {metadata}")

    def update_focus(self, wid=0) -> None:
        if self.focused == wid:
            return
        self.focused = wid

        def recheck_focus(_elapsed) -> None:
            if self.focused == wid:
                self.send("focus", wid, ())
        clock.schedule_once(recheck_focus, 0.01)


def make_client() -> XpraPygletClient:
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    client = XpraPygletClient()
    return client


def run_client(host: str, port: int) -> int:
    client = make_client()
    client.connect(host, port)
    return client.run()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable, Sequence, Iterable

from xpra.platform.paths import get_icon_filename
from xpra.scripts.parsing import audio_option
from xpra.net.common import PacketType
from xpra.net.compression import Compressed
from xpra.net.protocol.constants import CONNECTION_LOST
from xpra.common import FULL_INFO, noop, SizedBuffer
from xpra.os_util import get_machine_id, get_user_uuid, gi_import, OSX, POSIX
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, bytestostr, memoryview_to_bytes
from xpra.util.env import envint
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

avsynclog = Logger("av-sync")
log = Logger("client", "audio")

GLib = gi_import("GLib")

AV_SYNC_DELTA = envint("XPRA_AV_SYNC_DELTA")
DELTA_THRESHOLD = envint("XPRA_AV_SYNC_DELTA_THRESHOLD", 40)
DEFAULT_AV_SYNC_DELAY = envint("XPRA_DEFAULT_AV_SYNC_DELAY", 150)


def init_audio_tagging(tray_icon) -> None:
    if not POSIX:
        return
    try:
        from xpra import audio
        assert audio
    except ImportError:
        log("no audio module, skipping pulseaudio tagging setup")
        return
    try:
        from xpra.audio.pulseaudio.util import set_icon_path
        tray_icon_filename = get_icon_filename(tray_icon or "xpra")
        set_icon_path(tray_icon_filename)
    except ImportError as e:
        if not OSX:
            log.warn("Warning: failed to set pulseaudio tagging icon:")
            log.warn(" %s", e)


def get_matching_codecs(local_codecs, server_codecs) -> Sequence[str]:
    matching_codecs = tuple(x for x in local_codecs if x in server_codecs)
    log("get_matching_codecs(%s, %s)=%s", local_codecs, server_codecs, matching_codecs)
    return matching_codecs


def nooptions(*_args) -> Sequence[str]:
    return ()


class AudioClient(StubClientMixin):
    """
    Utility mixin for clients that handle audio
    """
    __signals__ = ["speaker-changed", "microphone-changed"]
    PREFIX = "audio"

    def __init__(self):
        self.audio_source_plugin = ""
        self.speaker_allowed: bool = False
        self.speaker_enabled: bool = False
        self.speaker_codecs = []
        self.microphone_allowed: bool = False
        self.microphone_enabled: bool = False
        self.microphone_codecs = []
        self.microphone_device = ""
        self.av_sync: bool = False
        self.av_sync_delta: int = AV_SYNC_DELTA
        self.audio_properties: typedict = typedict()
        # audio state:
        self.on_sink_ready: Callable[[], None] = noop
        self.audio_sink = None
        self.audio_sink_sequence: int = 0
        self.server_audio_eos_sequence: bool = False
        self.audio_source = None
        self.audio_source_sequence: int = 0
        self.audio_in_bytecount: int = 0
        self.audio_out_bytecount: int = 0
        self.audio_resume_restart = False
        self.server_av_sync: bool = False
        self.server_pulseaudio_id = ""
        self.server_pulseaudio_server = ""
        self.server_audio_decoders: Sequence[str] = ()
        self.server_audio_encoders: Sequence[str] = ()
        self.server_audio_receive: bool = False
        self.server_audio_send: bool = False
        self.queue_used_sent: int = 0
        # duplicated from ServerInfo mixin:
        self._remote_machine_id = ""

    def init(self, opts) -> None:
        self.av_sync = opts.av_sync
        self.speaker_allowed = audio_option(opts.speaker) in ("on", "off")
        # ie: "on", "off", "on:Some Device", "off:Some Device"
        mic = [x.strip() for x in opts.microphone.split(":", 1)]
        self.microphone_allowed = audio_option(mic[0]) in ("on", "off")
        self.microphone_device = ""
        if self.microphone_allowed and len(mic) == 2:
            self.microphone_device = mic[1]
        self.audio_source_plugin = opts.audio_source

        audio_option_fn: Callable = nooptions
        if self.speaker_allowed or self.microphone_allowed:
            def noaudio(title: str, message: str) -> None:
                self.may_notify_audio(title, message)
                self.speaker_allowed = False
                self.microphone_allowed = False
            try:
                from xpra.audio import common
                assert common
            except ImportError:
                noaudio("No Audio",
                        "`xpra-audio` subsystem is not installed\n"
                        " speaker and microphone forwarding are disabled")
                return
            try:
                from xpra.audio.common import audio_option_or_all
                audio_option_fn = audio_option_or_all
                from xpra.audio.wrapper import query_audio
                self.audio_properties = query_audio()
                if not self.audio_properties:
                    noaudio("No Audio",
                            "Audio subsystem query failed, is GStreamer installed?")
                    return
                gstv = self.audio_properties.strtupleget("gst.version")
                if gstv:
                    log.info("GStreamer version %s", ".".join(gstv[:3]))
                else:
                    log.info("GStreamer loaded")
            except Exception as e:
                log("failed to query audio", exc_info=True)
                noaudio("No Audio",
                        "Error querying the audio subsystem:\n"
                        f"{e}")
                return
        encoders = self.audio_properties.strtupleget("encoders")
        decoders = self.audio_properties.strtupleget("decoders")
        self.speaker_codecs = audio_option_fn("speaker-codec", opts.speaker_codec, decoders)
        self.microphone_codecs = audio_option_fn("microphone-codec", opts.microphone_codec, encoders)
        if not self.speaker_codecs:
            self.speaker_allowed = False
        if not self.microphone_codecs:
            self.microphone_allowed = False
        self.speaker_enabled = self.speaker_allowed and audio_option(opts.speaker) == "on"
        self.microphone_enabled = self.microphone_allowed and audio_option(mic[0]) == "on"
        log("speaker: codecs=%s, allowed=%s, enabled=%s", encoders, self.speaker_allowed, csv(self.speaker_codecs))
        log("microphone: codecs=%s, allowed=%s, enabled=%s, default device=%s",
            decoders, self.microphone_allowed, csv(self.microphone_codecs), self.microphone_device)
        log("av-sync=%s", self.av_sync)
        if POSIX and not OSX:
            try:
                from xpra.audio.pulseaudio.util import get_info as get_pa_info
                pa_info = get_pa_info()
                log("pulseaudio info=%s", pa_info)
                self.audio_properties.update(pa_info)
            except ImportError as e:
                log.warn("Warning: no pulseaudio information available")
                log.warn(" %s", e)
            except Exception:
                log.error("Error: failed to add pulseaudio info", exc_info=True)
        # audio tagging:
        init_audio_tagging(opts.tray_icon)

    def cleanup(self) -> None:
        self.stop_all_audio()

    def stop_all_audio(self) -> None:
        if self.audio_source:
            self.stop_sending_audio()
        if self.audio_sink:
            self.stop_receiving_audio()

    def get_info(self) -> dict[str, Any]:
        info: dict[str, Any] = {
            "speaker": self.speaker_enabled,
            "microphone": self.microphone_enabled,
            "properties": dict(self.audio_properties),
        }
        ss = self.audio_source
        if ss:
            info["src"] = ss.get_info()
        ss = self.audio_sink
        if ss:
            info["sink"] = ss.get_info()
        return {AudioClient.PREFIX: info}

    def get_caps(self) -> dict[str, Any]:
        return {
            "av-sync": self.get_avsync_capabilities(),
            AudioClient.PREFIX: self.get_audio_capabilities(),
        }

    def get_audio_capabilities(self) -> dict[str, Any]:
        if not self.audio_properties:
            return {}
        caps: dict[str, Any] = {
            "decoders": self.speaker_codecs,
            "encoders": self.microphone_codecs,
            "send": self.microphone_allowed,
            "receive": self.speaker_allowed,
        }
        # make mypy happy about the type: convert typedict to dict with string keys
        sp: dict[str, Any] = {str(k): v for k, v in self.audio_properties.items()}
        if FULL_INFO < 2:
            # only expose these specific keys:
            sp = {k: v for k, v in sp.items() if k in (
                "encoders", "decoders", "muxers", "demuxers",
            )}
        caps.update(sp)
        log("audio capabilities: %s", caps)
        return caps

    def get_avsync_capabilities(self) -> dict[str, Any]:
        if not self.av_sync:
            return {}
        delay = max(0, DEFAULT_AV_SYNC_DELAY + AV_SYNC_DELTA)
        return {
            "": True,
            "enabled": True,
            "delay.default": delay,
            "delay": delay,
        }

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.server_av_sync = c.boolget("av-sync.enabled")
        avsynclog("av-sync: server=%s, client=%s", self.server_av_sync, self.av_sync)
        audio = typedict(c.dictget("audio") or {})
        self.server_pulseaudio_id = audio.strget("pulseaudio.id")
        self.server_pulseaudio_server = audio.strget("pulseaudio.server")
        self.server_audio_decoders = audio.strtupleget("decoders")
        self.server_audio_encoders = audio.strtupleget("encoders")
        self.server_audio_receive = audio.boolget("receive")
        self.server_audio_send = audio.boolget("send")
        log("pulseaudio id=%s, server=%s, audio decoders=%s, audio encoders=%s, receive=%s, send=%s",
            self.server_pulseaudio_id, self.server_pulseaudio_server,
            csv(self.server_audio_decoders), csv(self.server_audio_encoders),
            self.server_audio_receive, self.server_audio_send)
        if self.server_audio_send and self.speaker_enabled:
            self.show_progress(90, "starting speaker forwarding")
            self.start_receiving_audio()
        if self.server_audio_receive and self.microphone_enabled:
            # call via idle_add because we may query X11 properties
            # to find the pulseaudio server:
            GLib.idle_add(self.start_sending_audio)
        return True

    def suspend(self) -> None:
        self.audio_resume_restart = bool(self.audio_sink)
        self.stop_receiving_audio()
        self.stop_sending_audio()

    def resume(self) -> None:
        ars = self.audio_resume_restart
        if ars:
            self.audio_resume_restart = False
            self.start_receiving_audio()

    ######################################################################
    # audio:

    def may_notify_audio(self, summary: str, body: str) -> None:
        # overridden in UI client subclass
        pass

    def audio_loop_check(self, mode="speaker") -> bool:
        from xpra.audio.gstreamer_util import ALLOW_SOUND_LOOP, loop_warning_messages
        if ALLOW_SOUND_LOOP:
            return True
        if self._remote_machine_id:
            if self._remote_machine_id != get_machine_id():
                # not the same machine, so OK
                return True
            if self._remote_uuid != get_user_uuid():
                # different user, assume different pulseaudio server
                return True
        # check pulseaudio id if we have it
        pulseaudio_id = self.audio_properties.get("pulseaudio", {}).get("id")
        if not pulseaudio_id or not self.server_pulseaudio_id:
            # not available, assume no pulseaudio so no loop?
            return True
        if self.server_pulseaudio_id != pulseaudio_id:
            # different pulseaudio server
            return True
        msgs = loop_warning_messages(mode)
        summary = msgs[0]
        body = "\n".join(msgs[1:])
        self.may_notify_audio(summary, body)
        log.warn("Warning: %s", summary)
        for x in msgs[1:]:
            log.warn(" %s", x)
        return False

    def no_matching_codec_error(self, forwarding="speaker",
                                server_codecs: Iterable[str] = (), client_codecs: Iterable[str] = ()) -> None:
        summary = "Failed to start %s forwarding" % forwarding
        body = "No matching codecs between client and server"
        self.may_notify_audio(summary, body)
        log.error("Error: %s", summary)
        log.error(" server supports: %s", csv(server_codecs))
        log.error(" client supports: %s", csv(client_codecs))

    def start_sending_audio(self, device="") -> None:
        """ (re)start an audio source and emit client signal """
        log("start_sending_audio(%s)", device)
        enabled = False
        try:
            assert self.microphone_allowed, "microphone forwarding is disabled"
            assert self.server_audio_receive, "client support for receiving audio is disabled"
            if not self.audio_loop_check("microphone"):
                return
            ss = self.audio_source
            if ss:
                enabled = True
                if ss.get_state() == "active":
                    log.error("Error: microphone forwarding is already active")
                else:
                    ss.start()
            else:
                enabled = self.start_audio_source(device)
        finally:
            if enabled != self.microphone_enabled:
                self.microphone_enabled = enabled
                self.emit("microphone-changed")
            log("start_sending_audio(%s) done, microphone_enabled=%s", device, enabled)

    def start_audio_source(self, device="") -> bool:
        log("start_audio_source(%s)", device)
        assert self.audio_source is None

        def audio_source_state_changed(*_args) -> None:
            self.emit("microphone-changed")

        # find the matching codecs:
        matching_codecs = get_matching_codecs(self.microphone_codecs, self.server_audio_decoders)
        log("start_audio_source(%s) matching codecs: %s", device, csv(matching_codecs))
        if not matching_codecs:
            self.no_matching_codec_error("microphone", self.server_audio_decoders, self.microphone_codecs)
            return False
        try:
            from xpra.audio.wrapper import start_sending_audio
            plugins = self.audio_properties.get("sources")
            ss = start_sending_audio(plugins, self.audio_source_plugin, device or self.microphone_device,
                                     "", 1.0, False, matching_codecs,
                                     self.server_pulseaudio_server, self.server_pulseaudio_id)
            if not ss:
                return False
            self.audio_source = ss
            ss.sequence = self.audio_source_sequence
            ss.connect("new-buffer", self.new_audio_buffer)
            ss.connect("state-changed", audio_source_state_changed)
            ss.connect("new-stream", self.new_stream)
            ss.start()
            log("start_audio_source(%s) audio source %s started", device, ss)
            return True
        except Exception as e:
            self.may_notify_audio("Failed to start microphone forwarding", "%s" % e)
            log.error("Error setting up microphone forwarding:")
            log.estr(e)
            return False

    def new_stream(self, audio_source, codec: str) -> None:
        log("new_stream(%s)", codec)
        if self.audio_source != audio_source:
            log("dropping new-stream signal (current source=%s, signal source=%s)", self.audio_source, audio_source)
            return
        codec = codec or audio_source.codec
        audio_source.codec = codec
        # tell the server this is the start:
        self.send("sound-data", codec, b"",
                  {
                      "start-of-stream": True,
                      "codec": codec,
                  })

    def stop_sending_audio(self) -> None:
        """ stop the audio source and emit client signal """
        ss = self.audio_source
        log("stop_sending_audio() audio source=%s", ss)
        if self.microphone_enabled:
            self.microphone_enabled = False
            self.emit("microphone-changed")
        self.audio_source = None
        if ss is None:
            log.warn("Warning: cannot stop audio capture which has not been started")
            return
        # tell the server to stop:
        self.send("sound-data", ss.codec or "", b"", {
            "end-of-stream": True,
            "sequence": ss.sequence,
        })
        self.audio_source_sequence += 1
        ss.cleanup()

    def start_receiving_audio(self) -> None:
        """ ask the server to start sending audio and emit the client signal """
        log("start_receiving_audio() audio sink=%s", self.audio_sink)
        enabled = False
        try:
            if self.audio_sink is not None:
                log("start_receiving_audio: we already have an audio sink")
                enabled = True
                return
            if not self.server_audio_send:
                log.error("Error receiving audio: support not enabled on the server")
                return
            if not self.audio_loop_check("speaker"):
                return
            # choose a codec:
            matching_codecs = get_matching_codecs(self.speaker_codecs, self.server_audio_encoders)
            log("start_receiving_audio() matching codecs: %s", csv(matching_codecs))
            if not matching_codecs:
                self.no_matching_codec_error("speaker", self.server_audio_encoders, self.speaker_codecs)
                return
            codec = matching_codecs[0]

            def sink_ready(*args) -> bool:
                scodec = codec
                log("sink_ready(%s) codec=%s (server codec name=%s)", args, codec, scodec)
                self.send("sound-control", "start", scodec)
                return False

            self.on_sink_ready = sink_ready
            enabled = self.start_audio_sink(codec)
        finally:
            if self.speaker_enabled != enabled:
                self.speaker_enabled = enabled
                self.emit("speaker-changed")
            log("start_receiving_audio() done, speaker_enabled=%s", enabled)

    def stop_receiving_audio(self, tell_server: bool = True) -> None:
        """
            ask the server to stop sending audio
            and toggle the flag so that we ignore further packets
            and emit the `new-sequence` client signal
        """
        ss = self.audio_sink
        log("stop_receiving_audio(%s) audio sink=%s", tell_server, ss)
        if self.speaker_enabled:
            self.speaker_enabled = False
            self.emit("speaker-changed")
        if not ss:
            return
        if tell_server and ss.sequence == self.audio_sink_sequence:
            self.send("sound-control", "stop", self.audio_sink_sequence)
        self.audio_sink_sequence += 1
        self.send("sound-control", "new-sequence", self.audio_sink_sequence)
        self.audio_sink = None
        log("stop_receiving_audio(%s) calling %s", tell_server, ss.cleanup)
        ss.cleanup()
        log("stop_receiving_audio(%s) done", tell_server)

    def audio_sink_state_changed(self, audio_sink, state: str) -> None:
        if audio_sink != self.audio_sink:
            log("audio_sink_state_changed(%s, %s) not the current sink, ignoring it", audio_sink, state)
            return
        log("audio_sink_state_changed(%s, %s) on_sink_ready=%s", audio_sink, state, self.on_sink_ready)
        if state == "ready":
            self.on_sink_ready()
            self.on_sink_ready = noop
        self.emit("speaker-changed")

    def audio_sink_bitrate_changed(self, audio_sink, bitrate: int) -> None:
        if audio_sink != self.audio_sink:
            log("audio_sink_bitrate_changed(%s, %s) not the current sink, ignoring it", audio_sink, bitrate)
            return
        log("audio_sink_bitrate_changed(%s, %s)", audio_sink, bitrate)
        # not shown in the UI, so don't bother with emitting a signal:
        # self.emit("speaker-changed")

    def audio_sink_error(self, audio_sink, error) -> None:
        log("audio_sink_error(%s, %s) exit_code=%s, current sink=%s",
            audio_sink, error, self.exit_code, self.audio_sink)
        if self.exit_code is not None:
            # exiting
            return
        if audio_sink != self.audio_sink:
            log("audio_sink_error(%s, %s) not the current sink, ignoring it", audio_sink, error)
            return
        estr = bytestostr(error).replace("gst-resource-error-quark: ", "")
        self.may_notify_audio("Speaker forwarding error", estr)
        log.warn("Error: stopping speaker:")
        log.warn(" %s", estr)
        self.stop_receiving_audio()

    def audio_process_stopped(self, audio_sink, *args) -> None:
        if self.exit_code is not None:
            # exiting
            return
        if audio_sink != self.audio_sink:
            log("audio_process_stopped(%s, %s) not the current sink, ignoring it", audio_sink, args)
            return
        log.warn("Warning: the audio process has stopped")
        self.stop_receiving_audio()

    def audio_sink_exit(self, audio_sink, *args) -> None:
        log("audio_sink_exit(%s, %s) audio_sink=%s", audio_sink, args, self.audio_sink)
        if self.exit_code is not None:
            # exiting
            return
        ss = self.audio_sink
        if audio_sink != ss:
            log("audio_sink_exit() not the current sink, ignoring it")
            return
        if ss and ss.codec:
            # the mandatory "I've been naughty warning":
            # we use the "codec" field as guard to ensure we only print this warning once…
            log.warn("Warning: the %s audio sink has stopped", ss.codec)
            ss.codec = ""
        self.stop_receiving_audio()

    def start_audio_sink(self, codec: str) -> bool:
        log("start_audio_sink(%s)", codec)
        assert self.audio_sink is None, "audio sink already exists!"
        try:
            log("starting %s audio sink", codec)
            from xpra.audio.wrapper import start_receiving_audio
            ss = start_receiving_audio(codec)
            if not ss:
                return False
            ss.sequence = self.audio_sink_sequence
            self.audio_sink = ss
            ss.connect("state-changed", self.audio_sink_state_changed)
            ss.connect("error", self.audio_sink_error)
            ss.connect("exit", self.audio_sink_exit)
            ss.connect(CONNECTION_LOST, self.audio_process_stopped)
            ss.start()
            log("%s audio sink started", codec)
            return True
        except Exception as e:
            log.error("Error: failed to start audio sink", exc_info=True)
            self.audio_sink_error(self.audio_sink, e)
            return False

    def new_audio_buffer(self, audio_source, data: bytes,
                         metadata: dict, packet_metadata: Sequence[SizedBuffer] = ()) -> None:
        log("new_audio_buffer(%s, %s, %s, %s)", audio_source, len(data or ()), metadata, packet_metadata)
        if audio_source.sequence < self.audio_source_sequence:
            log("audio buffer dropped: old sequence number: %s (current is %s)",
                audio_source.sequence, self.audio_source_sequence)
            return
        self.audio_out_bytecount += len(data)
        for x in packet_metadata:
            self.audio_out_bytecount += len(x)
        metadata["sequence"] = audio_source.sequence
        self.send_audio_data(audio_source, data, metadata, packet_metadata)

    def send_audio_data(self, audio_source, data: bytes,
                        metadata: dict, packet_metadata: Sequence[SizedBuffer]) -> None:
        codec = audio_source.codec
        # tag the packet metadata as already compressed:
        pmetadata = Compressed("packet metadata", packet_metadata)
        packet_data = [codec, Compressed(codec, data), metadata, pmetadata]
        self.send("sound-data", *packet_data)

    def send_audio_sync(self, v: int) -> None:
        self.send("sound-control", "sync", v)

    ######################################################################
    # packet handlers

    def _process_audio_data(self, packet: PacketType) -> None:
        codec = str(packet[1])
        data = memoryview_to_bytes(packet[2])
        metadata = typedict(packet[3])
        # the server may send packet_metadata, which is pushed before the actual audio data:
        packet_metadata = ()
        if len(packet) > 4:
            packet_metadata = packet[4]
        if data:
            self.audio_in_bytecount += len(data)
        # verify sequence number if present:
        seq = metadata.intget("sequence", -1)
        if self.audio_sink_sequence > 0 and 0 <= seq < self.audio_sink_sequence:
            log("ignoring audio data with old sequence number %s (now on %s)", seq, self.audio_sink_sequence)
            return

        if not self.speaker_enabled:
            if metadata.boolget("start-of-stream"):
                # server is asking us to start playing audio
                if not self.speaker_allowed:
                    # no can do!
                    log.warn("Warning: cannot honour the request to start the speaker")
                    log.warn(" speaker forwarding is disabled")
                    self.stop_receiving_audio(True)
                    return
                self.speaker_enabled = True
                self.emit("speaker-changed")
                self.on_sink_ready = noop
                codec = metadata.strget("codec")
                log("starting speaker on server request using codec %s", codec)
                self.start_audio_sink(codec)
            else:
                log("speaker is now disabled - dropping packet")
                return
        ss = self.audio_sink
        if ss is None:
            log("no audio sink to process audio data, dropping it")
            return
        if metadata.boolget("end-of-stream"):
            log("server sent end-of-stream for sequence %s, closing audio pipeline", seq)
            self.stop_receiving_audio(False)
            return
        if codec != ss.codec:
            log.error("Error: audio codec change is not supported!")
            log.error(" stream tried to switch from %s to %s", ss.codec, codec)
            self.stop_receiving_audio()
            return
        if ss.get_state() == "stopped":
            log("audio data received, audio sink is stopped - telling server to stop")
            self.stop_receiving_audio()
            return
        # (some packets (ie: sos, eos) only contain metadata)
        if data or packet_metadata:
            ss.add_data(data, dict(metadata), packet_metadata)
        if self.av_sync and self.server_av_sync:
            qinfo = typedict(ss.get_info()).dictget("queue")
            queue_used = typedict(qinfo or {}).intget("cur", -1)
            if queue_used < 0:
                return
            delta = (self.queue_used_sent or 0) - queue_used
            # avsynclog("server audio sync: queue info=%s, last sent=%s, delta=%s",
            #    dict((k,v) for (k,v) in info.items() if k.startswith("queue")), self.queue_used_sent, delta)
            if self.queue_used_sent is None or abs(delta) >= DELTA_THRESHOLD:
                avsynclog("server audio sync: sending updated queue.used=%i (was %s)",
                          queue_used, (self.queue_used_sent or "unset"))
                self.queue_used_sent = queue_used
                v = queue_used + self.av_sync_delta
                if self.av_sync_delta:
                    avsynclog(" adjusted value=%i with sync delta=%i", v, self.av_sync_delta)
                self.send_audio_sync(v)

    def init_authenticated_packet_handlers(self) -> None:
        log("init_authenticated_packet_handlers()")
        # this handler can run directly from the network thread:
        self.add_packets(f"{AudioClient.PREFIX}-data")
        self.add_legacy_alias("sound-data", f"{AudioClient.PREFIX}-data")



#!/usr/bin/env python3
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import get_native_tray_menu_helper_class, get_native_tray_classes
from xpra.platform.paths import get_icon_filename
from xpra.gtk.widget import scaled_image
from xpra.common import noop
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")
GdkPixbuf = gi_import("GdkPixbuf")

log = Logger("client")


class FakeApplication:

    def __init__(self):
        self.display_desc = {}
        self.session_name = "Test System Tray"
        self.windows_enabled = True
        self.readonly = False
        self.opengl_enabled = False
        self.modal_windows = False
        self.server_bell = False
        self.server_cursors = False
        self.server_readonly = False
        self.server_client_shutdown = True
        self.server_sharing = True
        self.server_sharing_toggle = True
        self.server_lock = True
        self.server_lock_toggle = True
        self.server_av_sync = True
        self.server_virtual_video_devices = 4
        self.server_webcam = True
        self.server_audio_send = True
        self.server_audio_receive = True
        self.server_clipboard = False
        self.server_encodings = ["png", "rgb"]
        self.server_encodings_with_quality = []
        self.server_encodings_with_speed = []
        self.server_start_new_commands = True
        self.server_xdg_menu = False
        self.server_commands_info = None
        self.server_multi_monitors = False
        self.server_bandwidth_limit = 0
        self.server_monitors = {}
        self.bandwidth_limit = 0
        self.speaker_allowed = True
        self.speaker_enabled = True
        self.microphone_enabled = True
        self.microphone_allowed = True
        self.client_supports_opengl = True
        self.client_supports_notifications = True
        self.client_supports_system_tray = True
        self.client_supports_clipboard = True
        self.client_supports_cursors = True
        self.client_supports_bell = True
        self.client_supports_sharing = True
        self.client_lock = False
        self.download_server_log = None
        self.remote_file_transfer = True
        self.remote_file_transfer_ask = True
        self.notifications_enabled = False
        self.client_clipboard_direction = "both"
        self.clipboard_enabled = True
        self.cursors_enabled = True
        self.default_cursor_data = None
        self.bell_enabled = False
        self.keyboard_helper = None
        self.av_sync = True
        self.webcam_forwarding = True
        self.webcam_device = None
        self.can_scale = True
        self.xscale = 1.0
        self.yscale = 1.0
        self.quality = 80
        self.speed = 50
        self.encoding = "png"
        self.send_download_request = None
        self._remote_subcommands = ()
        self._process_encodings = noop
        self._window_to_id = {}
        self.remote_printing_ask = False
        self.remote_open_files_ask = False
        self.remote_open_url_ask = False
        self._remote_server_log = ""
        self.start_new_commands = False
        classes = [get_native_tray_menu_helper_class()]
        try:
            from xpra.client.gtk3.tray_menu import GTKTrayMenu
            classes.append(GTKTrayMenu)
        except ImportError as e:
            log.warn("failed to load GTK tray menu class: %s", e)
        for hclass in classes:
            if hclass:
                try:
                    self.menu_helper = hclass(self)
                except Exception as e:
                    log.warn("failed to create menu helper %s: %s", hclass, e)
        assert self.menu_helper
        menu = self.menu_helper.build()
        tray_classes = list(get_native_tray_classes())
        try:
            from xpra.client.gtk3.statusicon_tray import GTKStatusIconTray
            tray_classes.append(GTKStatusIconTray)
        except ImportError:
            log("no StatusIcon tray")
        for tray_class in tray_classes:
            try:
                xpra_app_id = 0
                tray_icon_filename = "xpra"
                self.tray = tray_class(self, xpra_app_id, menu, "Test System Tray", tray_icon_filename,
                                       self.xpra_tray_geometry, self.xpra_tray_click,
                                       self.xpra_tray_mouseover, self.xpra_tray_exit)
            except Exception as e:
                log.warn("failed to create tray %s: %s", tray_class, e)
        self.tray.set_tooltip("Test System Tray")

    def after_handshake(self, cb: Callable, *args) -> None:
        GLib.idle_add(cb, *args)

    def on_server_setting_changed(self, setting: str, cb: Callable) -> None:
        """ this method is part of the GUI client "interface" """

    def connect(self, *args) -> None:
        """ this method is part of the GUI client "interface" """

    def get_encodings(self) -> Sequence[str]:
        from xpra.codecs.constants import PREFERRED_ENCODING_ORDER
        return PREFERRED_ENCODING_ORDER

    def show_start_new_command(self, *_args):
        """ this method is part of the GUI client "interface" """

    def show_server_commands(self, *_args):
        """ this method is part of the GUI client "interface" """

    def show_ask_data_dialog(self, *_args):
        """ this method is part of the GUI client "interface" """

    def show_file_upload(self, *_args):
        """ this method is part of the GUI client "interface" """

    def send_sharing_enabled(self, *_args):
        """ this method is part of the GUI client "interface" """

    def get_image(self, icon_name: str, size=None):
        with log.trap_error(f"Error loading image for icon {icon_name!r} and size {size}"):
            if not icon_name:
                return None
            icon_filename = get_icon_filename(icon_name)
            if not icon_filename:
                return None
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename=icon_filename)
            if not pixbuf:
                return None
            return scaled_image(pixbuf, size)

    def xpra_tray_click(self, button: int, pressed: bool, time: int = 0):
        log("xpra_tray_click(%s, %s, %s)", button, pressed, time)
        if button == 1 and pressed:
            GLib.idle_add(self.menu_helper.activate, button, time)
        elif button == 3 and not pressed:
            GLib.idle_add(self.menu_helper.popup, button, time)

    def xpra_tray_mouseover(self, *args):
        log("xpra_tray_mouseover(%s)", args)

    def xpra_tray_exit(self, *args):
        log("xpra_tray_exit%s", args)
        Gtk.main_quit()

    def xpra_tray_geometry(self, *args):
        log("xpra_tray_geometry%s geometry=%s", args, self.tray.get_geometry())

    def disconnect_and_quit(self, *_args):
        Gtk.main_quit()


def main() -> int:
    with program_context("tray", "Tray"):
        from xpra.util.system import is_X11
        if is_X11():
            from xpra.x11.gtk.display_source import init_gdk_display_source
            init_gdk_display_source()
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("tray test")
        FakeApplication()
        Gtk.main()
    return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import

GdkX11 = gi_import("GdkX11")


def GDKX11Window(*args, **kwargs) -> GdkX11.X11Window:
    # pylint: disable=import-outside-toplevel
    from xpra.gtk.window import new_GDKWindow
    return new_GDKWindow(GdkX11.X11Window, *args, **kwargs)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from unit.server.mixins.servermixintest_util import ServerMixinTest
from unit.process_test_util import DisplayContext


class DisplayMixinTest(ServerMixinTest):

    def test_display(self):
        with DisplayContext():
            self.do_test_display()

    def do_test_display(self):
        from xpra.server.mixins.display import DisplayManager
        from xpra.server.source.display import ClientDisplayMixin
        opts = AdHocStruct()
        opts.bell = True
        opts.cursors = True
        opts.dpi = 144
        opts.opengl = "no"
        opts.refresh_rate = "auto"

        def get_root_window_size() -> tuple[int, int]:
            return 1024, 768

        def calculate_workarea(*_args) -> None:
            pass

        def set_desktop_geometry(*_args) -> None:
            pass

        def make_display_manager():
            dm = DisplayManager()
            dm.get_root_window_size = get_root_window_size
            dm.calculate_workarea = calculate_workarea
            dm.set_desktop_geometry = set_desktop_geometry
            return dm
        self._test_mixin_class(make_display_manager, opts, {}, ClientDisplayMixin)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.util.str_fn import strtobytes, bytestostr
from xpra.util.io import get_util_logger


class AtomicInteger:
    __slots__ = ("counter", "lock")

    def __init__(self, integer: int = 0):
        self.counter: int = integer
        from threading import RLock
        self.lock: RLock = RLock()

    def increase(self, inc: int = 1) -> int:
        with self.lock:
            self.counter = self.counter + inc
            return self.counter

    def decrease(self, dec: int = 1) -> int:
        with self.lock:
            self.counter = self.counter - dec
            return self.counter

    def get(self) -> int:
        return self.counter

    def set(self, value: int):
        with self.lock:
            self.counter = value

    def __str__(self) -> str:
        return str(self.counter)

    def __repr__(self) -> str:
        return f"AtomicInteger({self.counter})"

    def __int__(self) -> int:
        return self.counter

    def __eq__(self, other) -> bool:
        try:
            return self.counter == int(other)
        except ValueError:
            return False

    def __cmp__(self, other) -> int:
        try:
            return self.counter - int(other)
        except ValueError:
            return -1


class MutableInteger:
    __slots__ = ("counter",)

    def __init__(self, integer: int = 0):
        self.counter: int = integer

    def increase(self, inc: int = 1) -> int:
        self.counter = self.counter + inc
        return self.counter

    def decrease(self, dec: int = 1) -> int:
        self.counter = self.counter - dec
        return self.counter

    def get(self) -> int:
        return self.counter

    def __str__(self) -> str:
        return str(self.counter)

    def __repr__(self) -> str:
        return f"MutableInteger({self.counter})"

    def __int__(self) -> int:
        return self.counter

    def __eq__(self, other) -> bool:
        return self.counter == int(other)

    def __ne__(self, other) -> bool:
        return self.counter != int(other)

    def __lt__(self, other) -> bool:
        return self.counter < int(other)

    def __le__(self, other) -> bool:
        return self.counter <= int(other)

    def __gt__(self, other) -> bool:
        return self.counter > int(other)

    def __ge__(self, other) -> bool:
        return self.counter >= int(other)

    def __cmp__(self, other) -> int:
        return self.counter - int(other)


# noinspection PyPep8Naming
class typedict(dict):
    __slots__ = ("warn",)  # no __dict__ - that would be redundant

    def __init__(self, mapping=(), **kwargs):
        super().__init__(mapping, **kwargs)
        self.warn = self._warn

    def get(self, k: str, default=None):
        if k in self:
            return super().get(k, default)
        # try to locate this value in a nested dictionary:
        if k.find(".") > 0:
            prefix, k = k.split(".", 1)
            if prefix in self:
                v = super().get(prefix)
                if isinstance(v, dict):
                    return typedict(v).get(k, default)
        return default

    def setdefault(self, k: str, default=None):
        return super().setdefault(k, default)

    def __repr__(self):
        return f'{type(self).__name__}({super().__repr__()})'

    def _warn(self, msg: str, *args):
        from xpra.log import Logger
        Logger("util").warn(msg, *args)

    def conv_get(self, key: str | int, default=None, conv=None):
        if key in self or not isinstance(key, str):
            v = super().get(key)
        else:
            # try harder by recursing:
            d = self
            while key.find(".") > 0:
                prefix, k = key.split(".", 1)
                if prefix not in d:
                    return default
                v = d[prefix]
                if not isinstance(v, dict):
                    return default
                d = v
                key = k
            if key not in d:
                return default
            v = dict.get(d, key)
        if isinstance(v, dict) and conv and conv in (bytestostr, strtobytes, int, bool):
            d = typedict(v)
            if "" in d:
                v = d[""]
        try:
            return conv(v)
        except (TypeError, ValueError, AssertionError) as e:
            self._warn(f"Warning: failed to convert {key!r}")
            self._warn(f" from {type(v)} using {conv}: {e}")
            return default

    def strget(self, k: str, default: str = "") -> str:
        return self.conv_get(k, default, bytestostr)

    def bytesget(self, k: str, default: bytes = b"") -> bytes:
        return self.conv_get(k, default, strtobytes)

    def intget(self, k: str, default: int = 0) -> int:
        return self.conv_get(k, default, int)

    def floatget(self, k: str, default: float = 0.0) -> float:
        return self.conv_get(k, default, float)

    def boolget(self, k: str, default: bool = False) -> bool:
        return self.conv_get(k, default, bool)

    def dictget(self, k: str | int, default: dict | None = None) -> dict:
        return self.conv_get(k, default, checkdict)

    def intpair(self, k: str, default_value: tuple[int, int] | None = None) -> tuple[int, int] | None:
        v = self.inttupleget(k, default_value)
        if v is None:
            return default_value
        if len(v) != 2:
            # "%s is not a pair of numbers: %s" % (k, len(v))
            return default_value
        try:
            return int(v[0]), int(v[1])
        except ValueError:
            return default_value

    def strtupleget(self, k: str, default_value=(),
                    min_items: int | None = None, max_items: int | None = None) -> tuple[str, ...]:
        return self.tupleget(k, default_value, str, min_items, max_items)

    def inttupleget(self, k: str, default_value=(),
                    min_items: int | None = None, max_items: int | None = None) -> tuple[int, ...]:
        return self.tupleget(k, default_value, int, min_items, max_items)

    def tupleget(self, k: str, default_value=(), item_type=None,
                 min_items: int | None = None, max_items: int | None = None) -> tuple[Any, ...]:
        v = self._listget(k, default_value, item_type, min_items, max_items)
        return tuple(v or ())

    def _listget(self, k: str, default_value, item_type=None,
                 min_items: int | None = None, max_items: int | None = None) -> list[Any] | tuple[Any, ...]:
        v = self.get(k)
        if v is None:
            return default_value
        if isinstance(v, dict) and "" in v:
            v = v.get("")
        if not isinstance(v, (list, tuple)):
            self._warn("listget%s", (k, default_value, item_type, max_items))
            self._warn("expected a list or tuple value for %s but got %s: %s", k, type(v), v)
            return default_value
        if min_items is not None and len(v) < min_items:
            self._warn("too few items in %s %s: minimum %s allowed, but got %s", type(v), k, max_items, len(v))
            return default_value
        if max_items is not None and len(v) > max_items:
            self._warn("too many items in %s %s: maximum %s allowed, but got %s", type(v), k, max_items, len(v))
            return default_value
        aslist = list(v)
        if item_type:
            for i, x in enumerate(aslist):
                if isinstance(x, bytes) and item_type is str:
                    x = bytestostr(x)
                    aslist[i] = x
                elif isinstance(x, str) and item_type is str:
                    x = str(x)
                    aslist[i] = x
                if not isinstance(x, item_type):
                    if callable(item_type):
                        try:
                            return item_type(x)
                        except Exception:
                            self._warn("invalid item type for %s %s: %s cannot be used with %s",
                                       type(v), k, item_type, type(x))
                            return default_value
                    self._warn("invalid item type for %s %s: expected %s but got %s",
                               type(v), k, item_type, type(x))
                    return default_value
        return aslist


# A simple little class whose instances we can stick random bags of attributes on.
class AdHocStruct:
    def __repr__(self):
        return ("<%s object, contents: %r>"
                % (type(self).__name__, self.__dict__))


def checkdict(v) -> dict:
    if isinstance(v, typedict):
        return dict(v)
    if not isinstance(v, dict):
        raise ValueError(f"expected a dict but got a {type(v)}")
    return v


def notypedict(d: dict, path="") -> dict:
    for k in list(d.keys()):
        v = d[k]
        if isinstance(v, typedict):
            log = get_util_logger()
            log.warn(f"Warning: found `typedict` at {path}.{k}")
            d[k] = notypedict(dict(v), f"{path}.{k}".strip("."))
        elif isinstance(v, dict):
            d[k] = notypedict(v, f"{path}.{k}".strip("."))
    return d


def make_instance(class_options, *args):
    log = get_util_logger()
    log("make_instance%s", tuple([class_options] + list(args)))
    for c in class_options:
        if c is None:
            continue
        try:
            v = c(*args)
            log(f"make_instance(..) {c}()={v}")
            if v:
                return v
        except Exception as e:
            log("make_instance(%s, %s)", class_options, args, exc_info=True)
            log.error("Error: cannot instantiate %s:", c)
            log.estr(e)
    return None


def reverse_dict(d: dict) -> dict:
    reversed_d = {}
    for k, v in d.items():
        reversed_d[v] = k
    return reversed_d


def merge_dicts(a: dict[str, Any], b: dict[str, Any], path: list[str] | None = None) -> dict[str, Any]:
    """ merges b into a """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                pathstr = ".".join(repr(v) for v in (path + [key]))
                # print(f"merged_dicts({a!r}, {b!r}, {path!r}) {key=!r}")
                raise ValueError(f"Conflict at {pathstr}: existing value is {a[key]}, new value is {b[key]}")
        else:
            a[key] = b[key]
    return a



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from math import log as mathlog, sqrt
from typing import Any

from xpra.server.cystats import (
    queue_inspect, logp, time_weighted_average,
    calculate_timesize_weighted_average_score,
)
from xpra.log import Logger

log = Logger("server", "stats")


def get_low_limit(mmap_enabled: bool, window_dimensions: tuple[int, int]) -> int:
    # the number of pixels which can be considered 'low' in terms of backlog.
    # Generally, just one full frame, (more with mmap because it is so fast)
    low_limit = 1024 * 1024
    ww, wh = window_dimensions
    if ww > 0 and wh > 0:
        low_limit = max(8 * 8, ww * wh)
    if mmap_enabled:
        # mmap can accumulate much more as it is much faster
        low_limit *= 4
    return low_limit


def calculate_batch_delay(wid: int, window_dimensions: tuple[int, int],
                          has_focus: bool, other_is_fullscreen: bool, other_is_maximized: bool, is_OR: bool,
                          soft_expired: int, batch, global_statistics, statistics,
                          bandwidth_limit: int, jitter: int) -> None:
    """
        Calculates a new batch delay.
        We first gather some statistics,
        then use them to calculate a number of factors.
        which are then used to adjust the batch delay in 'update_batch_delay'.
    """
    low_limit = get_low_limit(global_statistics.mmap_size > 0, window_dimensions)

    # for each indicator: (description, factor, weight)
    factors = statistics.get_factors(bandwidth_limit)
    statistics.target_latency = statistics.get_target_client_latency(global_statistics.min_client_latency,
                                                                     global_statistics.avg_client_latency,
                                                                     jitter=jitter)
    factors += global_statistics.get_factors(low_limit)
    # damage pixels waiting in the packet queue: (extract data for our window id only)
    time_values = global_statistics.get_damage_pixels(wid)

    def mayaddfac(metric: str, info: dict, factor: float, weight: float) -> None:
        if weight > 0.01:
            factors.append((metric, info, factor, weight))

    mayaddfac(*queue_inspect("damage-packet-queue-pixels", time_values, div=low_limit, smoothing=sqrt))
    # boost window that has focus and OR windows:
    mayaddfac("focus", {"has_focus": has_focus}, int(not has_focus), int(has_focus))
    mayaddfac("override-redirect", {"is_OR": is_OR}, int(not is_OR), int(is_OR))
    # soft expired regions is a strong indicator of problems:
    # (0 for none, up to max_soft_expired which is 5)
    mayaddfac("soft-expired", {"count": soft_expired}, soft_expired, int(bool(soft_expired)))
    # now use those factors to drive the delay change:
    min_delay = 0
    if batch.always:
        min_delay = batch.min_delay
    # if another window is fullscreen or maximized,
    # make sure we don't use a very low delay (cap at 10fps)
    if other_is_fullscreen or other_is_maximized:
        min_delay = max(100, min_delay)
    elif not has_focus:
        # wayland clients never tell us about minimized windows,
        # so we have to be more aggressive with unfocused windows instead:
        min_delay = max(40, min_delay)
    update_batch_delay(batch, factors, min_delay)


def update_batch_delay(batch, factors: list[tuple[str, dict, float, float]], min_delay=0) -> None:
    """
        Given a list of factors of the form:
        [(description, factor, weight)]
        we calculate a new batch delay.
        We use a time-weighted average of previous delays as a starting value,
        then combine it with the new factors.
    """
    current_delay = batch.delay
    now = monotonic()
    tv, tw = 0.0, 0.0
    decay = max(1, logp(current_delay / batch.min_delay) / 5.0)
    max_delay = batch.max_delay
    for delays, d_weight in ((batch.last_delays, 0.25), (batch.last_actual_delays, 0.75)):
        delays = tuple(delays or ())
        # get the weighted average
        # older values matter less, we decay them according to how much we batch already
        # (older values matter more when we batch a lot)
        for when, delay in delays:
            # newer matter more:
            w = d_weight / (1.0 + ((now - when) / decay) ** 2)
            d = max(0, min(max_delay, delay))
            tv += d * w
            tw += w
    hist_w = tw
    for x in factors:
        if len(x) != 4:
            log.warn("invalid factor line: %s" % str(x))
        else:
            log("update_batch_delay: %-28s : %.2f,%.2f  %s", x[0], x[2], x[3], x[1])
    valid_factors = tuple(x for x in factors if x is not None and len(x) == 4)
    all_factors_weight = sum(vf[-1] for vf in valid_factors)
    if all_factors_weight == 0:
        log("update_batch_delay: no weights yet!")
        return
    for _, _, factor, weight in valid_factors:
        target_delay = max(0, min(max_delay, current_delay * factor))
        w = max(hist_w, 1) * weight / all_factors_weight
        tw += w
        tv += target_delay * w
    batch.delay = int(max(min_delay, min(max_delay, tv // tw)))
    try:
        last_actual_delay = batch.last_actual_delays[-1][-1]
    except IndexError:
        last_actual_delay = -1
    log("update_batch_delay: delay=%i (last actual delay: %s)", batch.delay, last_actual_delay)
    batch.last_updated = now
    batch.factors = valid_factors


def get_target_speed(window_dimensions: tuple[int, int], batch,
                     global_statistics, statistics, bandwidth_limit: int,
                     min_speed: int, speed_data) -> tuple[dict[str, Any], int, int]:
    low_limit = get_low_limit(global_statistics.mmap_size > 0, window_dimensions)
    # ***********************************************************
    # encoding speed:
    #    0    for highest compression/slower
    #    100  for lowest compression/fast
    # here we try to minimize damage-latency and client decoding speed

    # backlog factor:
    _, pixels_backlog = statistics.get_client_backlog()
    pb_ratio = pixels_backlog / low_limit
    pixels_bl_s = 100 - int(100 * logp(pb_ratio / 4))  # 4 frames behind or more -> compress more

    # megapixels per second:
    mpixels = low_limit / 1024.0 / 1024.0
    # for larger window sizes, we should be downscaling,
    # and don't want to wait too long for those anyway:
    ref_damage_latency = (10 + 25 * (1 + mathlog(max(mpixels, 1)))) / 1000.0

    adil = statistics.avg_damage_in_latency or 0
    # abs: try to never go higher than N times the reference latency:
    dam_lat_abs = max(0, (adil - ref_damage_latency)) / (ref_damage_latency * 3)

    if batch.locked:
        dam_lat_rel = 0
        dam_lat_s = 100
    else:
        # calculate a target latency and try to get close to it
        avg_delay = batch.delay
        delays = tuple(batch.last_actual_delays)
        if delays:
            # average recent actual delay:
            avg_delay = time_weighted_average(delays)
        # and average that with the current delay (which is lower or equal):
        frame_delay = max(10, int((avg_delay + batch.delay) // 2))
        # ensure we always spend at least as much time encoding as we spend batching:
        # (one frame encoding whilst one frame is batching is our ideal result)
        target_damage_latency = max(ref_damage_latency, frame_delay / 1000.0)
        dam_target_speed = min_speed
        if speed_data:
            dam_target_speed = max(min_speed, time_weighted_average(speed_data))
        # rel: do we need to increase speed to reach the target:
        dam_lat_rel = dam_target_speed / 100.0 * adil / target_damage_latency
        # cap the speed if we're delaying frames longer than we should:
        # (so we spend more of that time compressing them better instead):
        dam_lat_s = int(100 * 2 * ref_damage_latency * 1000 // frame_delay)

    # if we have more pixels to encode, we may need to go faster
    # (this is important because the damage latency used by the other factors
    # may aggregate multiple damage requests into one packet - which may skip frames)
    # TODO: reconcile this with video regions
    # only count the last second's worth:
    now = monotonic()
    lim = now - 1.0
    lde = tuple(w * h for t, _, _, w, h in tuple(statistics.last_damage_events) if t >= lim)
    pixels = sum(lde)
    mpixels_per_s = pixels / (1024 * 1024)
    pps = 0.0
    pixel_rate_s = 100
    if len(lde) > 5 and mpixels_per_s >= 1:
        # above 50 MPixels/s, we should reach 100% speed
        # (even x264 peaks at tens of MPixels/s)
        pps = sqrt(mpixels_per_s / 50.0)
        # if there aren't many pixels,
        # we can spend more time compressing them better:
        # (since it isn't going to cost too much to compress)
        # ie: 2MPixels/s -> max_speed=60%
        pixel_rate_s = 20 + int(mpixels_per_s * 20)

    bandwidth_s = 100
    if bandwidth_limit > 0:
        # below N Mbps, lower the speed ceiling,
        # so we will compress better:
        N = 10
        bandwidth_s = int(100 * sqrt(bandwidth_limit / (N * 1000 * 1000)))

    gcv = global_statistics.congestion_value
    congestion_s = 100
    if gcv > 0:
        # apply strict limit for congestion events:
        congestion_s = max(0, int(100 - gcv * 1000))

    # ensure we decode at a reasonable speed (for slow / low-power clients)
    # maybe this should be configurable?
    min_decode_speed = 1 * 1000 * 1000  # MPixels/s
    ads = statistics.avg_decode_speed or 0
    dec_lat = 0.0
    if ads > 0:
        dec_lat = min_decode_speed / ads

    ms = min(100, max(min_speed, 0))
    max_speed = max(ms, min(pixels_bl_s, dam_lat_s, pixel_rate_s, bandwidth_s, congestion_s))
    # combine factors: use the highest one:
    target = min(max(dam_lat_abs, dam_lat_rel, dec_lat, pps, 0), 1)
    # scale target between min_speed and 100:
    speed = int(ms + (100 - ms) * target)
    speed = max(ms, min(max_speed, speed))

    # expose data we used:
    info = {
        "low-limit": int(low_limit),
        "max-speed": int(max_speed),
        "min-speed": int(min_speed),
        "factors": {
            "damage-latency-abs": int(dam_lat_abs * 100),
            "damage-latency-rel": int(dam_lat_rel * 100),
            "decoding-latency": int(dec_lat * 100),
            "pixel-rate": int(pps * 100),
        },
        "limits": {
            "backlog": pixels_bl_s,
            "damage-latency": dam_lat_s,
            "pixel-rate": pixel_rate_s,
            "bandwidth-limit": bandwidth_s,
            "congestion": congestion_s,
        },
    }
    return info, round(speed), round(max_speed)


def get_target_quality(window_dimensions: tuple[int, int], batch,
                       global_statistics, statistics, bandwidth_limit,
                       min_quality: int, min_speed: int) -> tuple[dict[str, Any], int]:
    low_limit = get_low_limit(global_statistics.mmap_size > 0, window_dimensions)
    # ***********************************************************
    # quality:
    #    0    for lowest quality (low bandwidth usage)
    #    100  for best quality (high bandwidth usage)
    # here we try to minimize client-latency, packet-backlog and batch.delay
    # the compression ratio tells us if we can increase the quality

    # backlog factor:
    packets_backlog, pixels_backlog = statistics.get_client_backlog()
    pb_ratio = pixels_backlog / low_limit
    pixels_bl_q = 1 - sqrt(pb_ratio / 8)  # 8 frames behind or more -> min quality

    # bandwidth limit factor:
    bandwidth_q = 1
    if bandwidth_limit > 0:
        # below 10Mbps, lower the quality
        bandwidth_q = int(100 * sqrt(bandwidth_limit / (10.0 * 1000 * 1000)))

    # congestion factor:
    gcv = global_statistics.congestion_value
    congestion_q = 1 - gcv * 10

    # batch delay factor:
    batch_q = 1
    if batch is not None:
        recs = len(batch.last_actual_delays)
        if recs > 0 and not batch.locked:
            # weighted average between start delay and min_delay,
            # so when we start without any records, we don't lower quality
            # just because the start delay is higher than min_delay
            # anything less than N times the reference delay is good enough:
            N = 3.0 - min_speed / 50.0
            # if the min-speed is high, reduce tolerance:
            tolerance = 10 - int(min_speed // 10)
            ref_delay = max(0, tolerance + N * (batch.start_delay * 10 + batch.min_delay * recs) // (recs + 10))
            batch_q = (N * ref_delay) / max(1, batch.min_delay, batch.delay)

    # latency limit factor:
    latency_q = 1
    if global_statistics.client_latency and global_statistics.recent_client_latency > 0:
        # if the recent latency is too high, keep quality lower:
        latency_q = 3.0 * statistics.target_latency / global_statistics.recent_client_latency

    # target is the lowest value of all those limits:
    target = max(min(pixels_bl_q, bandwidth_q, congestion_q, batch_q, latency_q, 1), 0)

    info: dict[str, Any] = {}
    # boost based on recent compression ratio
    comp_boost = 0
    # from here on, the compression ratio integer value is in per-1000:
    es = tuple((t, pixels, 1000 * compressed_size * bpp // pixels // 32)
               for (t, _, pixels, bpp, compressed_size, _) in tuple(statistics.encoding_stats) if pixels >= 4096)
    if len(es) >= 2:
        # use the recent vs average compression ratio
        # (add value to smooth things out a bit, so very low compression ratios don't skew things)
        comp_boost = 0
        ascore, rscore = calculate_timesize_weighted_average_score(es)
        if ascore > rscore:
            # raise the quality
            # but only if there is no backlog:
            if packets_backlog == 0:
                smooth = 150
                comp_boost = logp((smooth + ascore) / (smooth + rscore)) - 1.0
        else:
            # lower the quality
            # more so if the compression is not doing very well:
            mult = (1000 + rscore) / 2000.0  # mult should be in the range 0.5 to ~1.0
            smooth = 50
            comp_boost = -logp(((smooth + rscore) / (smooth + ascore)) - 1.0) * mult
        info["compression-ratio"] = ascore, rscore
        target = max(0, target + comp_boost)

    # discount the quality more aggressively if we have speed requirements to satisfy:
    if min_speed > 0:
        # ie: for min_speed=50:
        # target=1.0   -> target=1.0
        # target=0.8   -> target=0.51
        # target=0.5   -> target=0.125
        # target=0     -> target=0
        target = target ** ((100.0 + 4 * min_speed) / 100.0)

    # raise the quality when there are not many recent damage events:
    ww, wh = window_dimensions
    if ww > 0 and wh > 0:
        lde = tuple(statistics.last_damage_events)
        if lde:
            now = monotonic()
            damage_pixel_count = tuple((lim, sum(w * h for t, _, _, w, h in lde if now - lim <= t < now - lim + 1))
                                       for lim in range(1, 11))
            pixl5 = sum(v for lim, v in damage_pixel_count if lim <= 5)
            pixn5 = sum(v for lim, v in damage_pixel_count if lim > 5)
            pctpixdamaged = pixl5 / (ww * wh)
            log("get_target_quality: target=%3i%% (window %4ix%-4i) pctpixdamaged=%3i%%, dpc=%s",
                100 * target, ww, wh, pctpixdamaged * 100, damage_pixel_count)
            if pctpixdamaged < 0.5:
                target *= (1.5 - pctpixdamaged)
            if pixl5 < pixn5:
                target = sqrt(target)

    # apply min-quality:
    mq = min(100, max(min_quality, 0))
    quality = int(mq + (100 - mq) * target)
    quality = max(0, mq, min(100, quality))

    info.update({
        "min-quality": min_quality,
        "min-speed": min_speed,
        "backlog": (packets_backlog, pixels_backlog, low_limit, int(100 * pb_ratio)),
        "limits": {
            "backlog": int(pixels_bl_q * 100),
            "bandwidth": int(bandwidth_q * 100),
            "congestion": int(congestion_q * 100),
            "batch": int(batch_q * 100),
            "latency": int(latency_q * 100),
            "boost": int(comp_boost * 100),
        },
    })
    return info, int(quality)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import unittest

from unit.test_util import silence_error
from xpra.util.objects import AdHocStruct
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.server.mixins import logging
from unit.server.mixins.servermixintest_util import ServerMixinTest


class nostr():
    def __str__(self):
        raise Exception("test format failure")


class InputMixinTest(ServerMixinTest):

    def test_logging(self):
        opts = AdHocStruct()
        opts.remote_logging = "yes"
        log_messages = []

        def FakeSource():
            s = StubSourceMixin()
            s.counter = 0
            return s

        def do_log(level, line):
            log_messages.append((level, line))

        def _LoggingServer():
            ls = logging.LoggingServer()
            ls.do_log = do_log
            return ls

        self._test_mixin_class(_LoggingServer, opts, {}, FakeSource)
        self.handle_packet(("logging", 10, "hello", time.time()))
        message = log_messages[0]
        assert message[0] == 10
        assert message[1].endswith("hello")
        #multi-part:
        self.handle_packet(("logging", 20, ["multi", "messages"], time.time()))
        #invalid:
        with silence_error(logging):
            self.handle_packet(("logging", 20, nostr(), time.time()))

    def test_invalid(self):
        l = logging.LoggingServer()
        opts = AdHocStruct()
        opts.remote_logging = "on"
        l.init(opts)
        l._process_logging(None, ())  # pylint: disable=protected-access


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.gtk.signals import register_os_signals
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label, modify_fg, color_parse
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.platform.gui import force_focus
from xpra.os_util import gi_import
from xpra.util.io import get_util_logger

Gtk = gi_import("Gtk")
GdkPixbuf = gi_import("GdkPixbuf")

log = get_util_logger()


class ConfirmDialogWindow(Gtk.Dialog):

    def __init__(self, title="Title", prompt="", info=(), icon="", buttons=()):
        log("ConfirmDialogWindow%s", (title, prompt, info, icon, buttons))
        super().__init__()
        self.set_border_width(20)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.connect("delete-event", self.quit)
        self.set_default_size(400, 150)
        self.set_title(title)
        add_close_accel(self, self.quit)

        icon_pixbuf = get_icon_pixbuf(icon)
        if icon_pixbuf:
            self.set_icon(icon_pixbuf)

        vbox = self.get_content_area()
        vbox.set_spacing(10)

        def al(text: str, font="sans 14", xalign=0.0) -> Gtk.Alignment:
            lbl = label(text, font=font)
            if text.startswith("WARNING"):
                modify_fg(lbl, color_parse("red"))
            align = Gtk.Alignment(xalign=xalign, yalign=0.5, xscale=0.0, yscale=0)
            align.add(lbl)
            align.show_all()
            return align

        vbox.add(al(title, "sans 18", 0.5))
        info_box = Gtk.VBox()
        for i in info:
            info_box.add(al(i))
        info_box.show_all()
        vbox.add(info_box)
        vbox.add(al(prompt))

        # Buttons:
        for txt, code in buttons:
            btn = self.add_button(txt, code)
            btn.set_size_request(100, 48)

    def quit(self, *args) -> bool:
        log("quit%s", args)
        self.destroy()
        return True

    def signal_quit(self, *args) -> None:
        log("signal_quit%s", args)
        self.close()


def show_confirm_dialog(argv) -> int:
    from xpra.platform.gui import ready as gui_ready, init as gui_init, set_default_icon

    set_default_icon("information.png")
    gui_init()

    log("show_confirm_dialog(%s)", argv)

    def arg(i: int) -> str:
        if len(argv) <= i:
            return ""
        return argv[i].replace("\\n\\r", "\\n").replace("\\n", "\n")

    title = arg(0) or "Confirm Key"
    prompt = arg(1)
    info = arg(2)
    icon = arg(3)
    buttons = []
    n = 4
    while len(argv) > (n + 1):
        text = arg(n)
        try:
            code = int(arg(n + 1))
        except ValueError as e:
            log.error("Error: confirm dialog cannot parse code '%s': %s", arg(n + 1), e)
            return 1
        buttons.append((text, code))
        n += 2
    app = ConfirmDialogWindow(title, prompt, info, icon, buttons)
    register_os_signals(app.signal_quit, "Dialog")
    gui_ready()
    force_focus()
    app.show_all()
    app.present()
    return app.run()


def main(args):
    from xpra.log import consume_verbose_argv
    from xpra.platform import program_context
    with program_context("Confirm-Dialog", "Confirm Dialog"):
        consume_verbose_argv(args, "util",)
        try:
            return show_confirm_dialog(args[1:])
        except KeyboardInterrupt:
            return 1


if __name__ == "__main__":
    v = main(sys.argv)
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.util.env import envint
from xpra.codecs.constants import LOSSY_PIXEL_FORMATS, CSCSpec, VideoSpec
from xpra.log import Logger

log = Logger("score")

GPU_BIAS = envint("XPRA_GPU_BIAS", 100)
MIN_FPS_COST = envint("XPRA_MIN_FPS_COST", 4)

# any colourspace conversion will lose at least some quality (due to rounding)
# (so add 0.2 to the value we get from calculating the degradation using get_subsampling_divs)
SUBSAMPLING_QUALITY_LOSS = {
    "NV12": 186,
    "YUV420P": 186,  # 1.66 + 0.2
    "YUV422P": 153,  # 1.33 + 0.2
    "YUV444P": 120,  # 1.00 + 0.2
}

LOSSY_CSC = ("NV12", "YUV420P", "YUV422P")


def clamp(value: int) -> int:
    return max(0, min(100, value))


def get_quality_score(csc_format: str, csc_spec: CSCSpec | None, encoder_spec: VideoSpec, scaling: tuple[int, int],
                      target_quality: int = 100, min_quality: int = 0) -> int:
    quality = encoder_spec.quality
    div = SUBSAMPLING_QUALITY_LOSS.get(csc_format, 100)
    quality = quality * 100 // div

    if csc_spec:
        # csc_spec.quality is the upper limit (up to 100):
        quality += csc_spec.quality
        quality /= 2.0

    if scaling == (1, 1) and csc_format not in LOSSY_CSC and target_quality == 100 and encoder_spec.has_lossless_mode:
        # we want lossless!
        qscore = quality + 80
    else:
        # how far are we from the current quality heuristics?
        qscore = 100 - abs(target_quality - quality)
        if min_quality >= quality:
            # if this encoder's quality is lower than the min_quality
            # then it isn't very suitable, discount its score:
            mqs = (min_quality - quality) // 2
            qscore = max(0, qscore - mqs)
        # when downscaling, YUV420P should always win:
        if csc_format in ("YUV420P", "NV12") and scaling != (1, 1):
            qscore *= 2.0
    return round(qscore)


def get_speed_score(csc_format: str, csc_spec: CSCSpec | None, encoder_spec: VideoSpec, scaling: tuple[int, int],
                    target_speed: int = 100, min_speed: int = 0) -> int:
    # when subsampling, add the speed gains to the video encoder
    # which now has less work to do:
    mult = {
        "NV12": 100,
        "YUV420P": 100,
        "YUV422P": 80,
    }.get(csc_format, 60)
    # score based on speed:
    speed = round(encoder_spec.speed * mult // 100)
    # the encoder speed matters less
    # when the target speed is low:
    ts = clamp(target_speed)
    sscore = speed * 100 // (200 - ts)
    if csc_spec:
        # if there is a csc step,
        # then we lose some performance,
        # but less if the csc is fast
        sscore = sscore - 20 - (100 - csc_spec.speed) // 2
    # when already downscaling, favour YUV420P subsampling:
    if csc_format in ("YUV420P", "NV12") and scaling != (1, 1):
        sscore += 25
    if min_speed >= speed:
        # if this encoder's speed is lower than the min_speed
        # then it isn't very suitable, discount its score:
        mss = (min_speed - speed) // 2
        sscore -= mss
    return round(sscore)


def get_pipeline_score(enc_in_format: str, csc_spec: CSCSpec | None, encoder_spec: VideoSpec,
                       width: int, height: int, scaling: tuple[int, int],
                       target_quality: int, min_quality: int,
                       target_speed: int, min_speed: int,
                       current_csce, current_ve,
                       score_delta: int, ffps: int, detection=True) -> tuple | None:
    """
        Given an optional csc step (csc_format and csc_spec),
        and a required encoding step (encoder_spec and width/height),
        we calculate a score of how well this matches our requirements:
        * our quality target "self._currend_quality"
        * our speed target "self._current_speed"
        * how expensive it would be to switch to this pipeline option
        Note: we know the current pipeline settings, so the "switching
        cost" will be lower for pipelines that share components with the
        current one.

        Can be called from any thread.
    """

    qscore = clamp(get_quality_score(enc_in_format, csc_spec, encoder_spec, scaling, target_quality, min_quality))
    sscore = get_speed_score(enc_in_format, csc_spec, encoder_spec, scaling, target_speed, min_speed)

    # multiplier for setup_cost:
    # (lose points if we have less than N fps)
    setup_cost_mult = int(detection) * (1 + max(0, MIN_FPS_COST - ffps))

    # how well the codec deals with larger screen sizes:
    sizescore = 100
    pixels = width * height
    if scaling != (1, 1):
        n, d = scaling
        pixels = pixels * n * n // d // d
    if pixels >= 1048576:
        # high size efficiency means sizescore stays high even with high number of mpixels,
        # ie: 1MPixels -> sizescore = 100
        # ie: 8MPixels -> sizescore = size_efficiency
        sdisc = 100 - encoder_spec.size_efficiency
        sizescore = max(0, 100 - pixels * sdisc // 1048576 // 4)

    # runtime codec adjustments:
    runtime_score = 100
    # score for "edge resistance" via setup cost:

    csc_width = 0
    csc_height = 0
    if csc_spec:
        # OR the masks, so we have a chance of making it work
        width_mask = csc_spec.width_mask & encoder_spec.width_mask
        height_mask = csc_spec.height_mask & encoder_spec.height_mask
        csc_width = width & width_mask
        csc_height = height & height_mask
        if enc_in_format == "RGB":
            # converting to "RGB" is often a waste of CPU
            # (can only get selected because the csc step will do scaling,
            # but even then, the YUV subsampling are better options)
            ecsc_score = 1
        else:
            ecsc_score = 80

            # if anything is different, account for setup cost:

            def is_csc_changed() -> bool:
                if current_csce is None:
                    return True
                if current_csce.get_dst_format() != enc_in_format:
                    return True
                if current_csce.get_type() != csc_spec.codec_type:
                    return True
                if current_csce.get_src_width() != csc_width or current_csce.get_src_height() != csc_height:
                    return True
                return False

            if is_csc_changed():
                # if we have to change csc, account for new csc setup cost:
                ecsc_score = max(0, 80 - round(csc_spec.setup_cost * setup_cost_mult * 80 // 100))
        ecsc_score += csc_spec.score_boost
        runtime_score *= csc_spec.get_runtime_factor()

        csc_scaling = scaling
        encoder_scaling = (1, 1)
        if scaling != (1, 1) and not csc_spec.can_scale:
            # csc cannot take care of scaling, so encoder will have to:
            encoder_scaling = scaling
            csc_scaling = (1, 1)
        if scaling != (1, 1):
            # if we are (down)scaling, we should prefer lossy pixel formats:
            v = LOSSY_PIXEL_FORMATS.get(enc_in_format, 1)
            qscore *= (v / 2)
        enc_width, enc_height = get_encoder_dimensions(encoder_spec, csc_width, csc_height, scaling)
    else:
        # not using csc at all!
        ecsc_score = 100
        width_mask = encoder_spec.width_mask
        height_mask = encoder_spec.height_mask
        enc_width = width & width_mask
        enc_height = height & height_mask
        csc_scaling = None
        encoder_scaling = scaling

    if encoder_scaling != (1, 1) and not encoder_spec.can_scale:
        # we need the encoder to scale but it cannot do it, fail it:
        log("scaling (%s) not supported by %s", encoder_scaling, encoder_spec)
        return None

    if enc_width < encoder_spec.min_w or enc_height < encoder_spec.min_h:
        log("video size %ix%i out of range for %s, min %ix%i",
            enc_width, enc_height, encoder_spec.codec_type, encoder_spec.min_w, encoder_spec.min_h)
        return None
    elif enc_width > encoder_spec.max_w or enc_height > encoder_spec.max_h:
        log("video size %ix%i out of range for %s, max %ix%i",
            enc_width, enc_height, encoder_spec.codec_type, encoder_spec.max_w, encoder_spec.max_h)
        return None

    ee_score = 100
    if current_ve is None or current_ve.get_type() != encoder_spec.codec_type or \
            current_ve.get_src_format() != enc_in_format or \
            current_ve.get_width() != enc_width or current_ve.get_height() != enc_height:
        # account for new encoder setup cost:
        ee_score = 100 - round(encoder_spec.setup_cost * setup_cost_mult)
        ee_score += encoder_spec.score_boost
    # edge resistance score: average of csc and encoder score:
    er_score = (ecsc_score + ee_score) // 2
    # gpu vs cpu
    gpu_score = max(0, GPU_BIAS - 50) * encoder_spec.gpu_cost // 50
    cpu_score = max(0, 50 - GPU_BIAS) * encoder_spec.cpu_cost // 50
    runtime_score *= encoder_spec.get_runtime_factor()
    score = round(
        (qscore + sscore + er_score + sizescore + score_delta + gpu_score + cpu_score) * runtime_score / 100)
    log(
        "get_pipeline_score(%-7s, %-24r, %-42r, %5i, %5i) quality: %3i, speed: %3i, setup: %4i - %4i runtime: %3i scaling: %s / %s, encoder dimensions=%sx%s, sizescore=%3i, client score delta=%3i, cpu score=%3i, gpu score=%3i, score=%3i",
        # noqa: E501
        enc_in_format, csc_spec, encoder_spec, width, height,
        qscore, sscore, ecsc_score, ee_score, runtime_score, scaling,
        encoder_scaling, enc_width, enc_height, sizescore, score_delta,
        cpu_score, gpu_score, score)
    return (
        score, scaling, csc_scaling, csc_width, csc_height, csc_spec, enc_in_format,
        encoder_scaling, enc_width, enc_height, encoder_spec,
    )


def get_encoder_dimensions(encoder_spec: VideoSpec, width: int, height: int, scaling=(1, 1)) -> tuple[int, int]:
    """
        Given a csc and encoder specs and dimensions, we calculate
        the dimensions that we would use as output.
        Taking into account:
        * applications can require scaling (see "scaling" attribute)
        * we scale fullscreen and maximize windows when at high speed
          and low quality.
        * we do not bother scaling small dimensions
        * the encoder may not support all dimensions
          (see width and height masks)
    """
    v, u = scaling
    enc_width = int(width * v / u) & encoder_spec.width_mask
    enc_height = int(height * v / u) & encoder_spec.height_mask
    return enc_width, enc_height



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False
from typing import Tuple, Dict
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.log import Logger
log = Logger("decoder", "spng")

from xpra.codecs.spng.spng cimport (
    SPNG_VERSION_MAJOR, SPNG_VERSION_MINOR, SPNG_VERSION_PATCH,
    SPNG_CTX_ENCODER,
    SPNG_DECODE_TRNS,
    SPNG_FMT_RGBA8, SPNG_FMT_RGBA16, SPNG_FMT_RGB8,
    SPNG_COLOR_TYPE_GRAYSCALE, SPNG_COLOR_TYPE_TRUECOLOR, SPNG_COLOR_TYPE_GRAYSCALE_ALPHA,
    SPNG_COLOR_TYPE_TRUECOLOR_ALPHA, SPNG_COLOR_TYPE_INDEXED,
    spng_ctx, spng_ihdr, spng_strerror,
    spng_ctx_new, spng_ctx_free,
    spng_get_ihdr, spng_format,
    spng_decode_image, spng_set_png_buffer, spng_decoded_image_size,
)
from libc.stdint cimport uintptr_t, uint32_t, uint8_t
from xpra.buffers.membuf cimport getbuf, MemBuf  # pylint: disable=syntax-error
from xpra.util.env import envint

MAX_SIZE = envint("XPRA_SPNG_MAX_SIZE", 8192*8192)


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS


COLOR_TYPE_STR: Dict[int, str] = {
    SPNG_COLOR_TYPE_GRAYSCALE       : "GRAYSCALE",
    SPNG_COLOR_TYPE_TRUECOLOR       : "TRUECOLOR",
    SPNG_COLOR_TYPE_INDEXED         : "INDEXED",
    SPNG_COLOR_TYPE_GRAYSCALE_ALPHA : "GRAYSCALE_ALPHA",
    SPNG_COLOR_TYPE_TRUECOLOR_ALPHA : "TRUECOLOR_ALPHA",
}


def get_version() -> Tuple[int, int, int]:
    return (SPNG_VERSION_MAJOR, SPNG_VERSION_MINOR, SPNG_VERSION_PATCH)


def get_encodings() -> Sequence[str]:
    return ("png", "png/L", "png/P")


def get_error_str(int r) -> str:
    s = spng_strerror(r)
    return str(s)


def check_error(int r, msg) -> int:
    if r:
        log_error(r, msg)
    return r


def log_error(int r, msg) -> None:
    log.error("Error: %s", msg)
    log.error(" code %i: %s", r, get_error_str(r))


def decompress(data, options: typedict) -> Tuple[memoryview, str, int, int]:
    cdef spng_ctx *ctx = spng_ctx_new(0)
    if ctx==NULL:
        raise RuntimeError("failed to instantiate an spng context")

    cdef Py_buffer py_buf
    if PyObject_GetBuffer(data, &py_buf, PyBUF_ANY_CONTIGUOUS):
        spng_ctx_free(ctx)
        raise ValueError(f"failed to read compressed data from {type(data)}")

    cdef int r
    def close():
        PyBuffer_Release(&py_buf)
        spng_ctx_free(ctx)

    if check_error(spng_set_png_buffer(ctx, py_buf.buf, py_buf.len),
                   "failed to set png buffer"):
        close()
        return None

    cdef spng_ihdr ihdr
    if check_error(spng_get_ihdr(ctx, &ihdr),
                   "failed to get ihdr"):
        close()
        return None

    log("ihdr: %ix%i-%i, color-type=%s, compression-method=%#x, filter-method=%#x, interlace-method=%#x",
        ihdr.width, ihdr.height, ihdr.bit_depth, COLOR_TYPE_STR.get(ihdr.color_type, ihdr.color_type),
        ihdr.compression_method, ihdr.filter_method, ihdr.interlace_method)

    cdef int flags = 0
    cdef size_t out_size
    cdef spng_format fmt
    if ihdr.color_type==SPNG_COLOR_TYPE_TRUECOLOR:
        fmt = SPNG_FMT_RGB8
        rgb_format = "RGB"
    elif ihdr.color_type==SPNG_COLOR_TYPE_TRUECOLOR_ALPHA:
        fmt = SPNG_FMT_RGBA8
        rgb_format = "RGBA"
        flags = SPNG_DECODE_TRNS
    elif ihdr.color_type==SPNG_COLOR_TYPE_GRAYSCALE:
        fmt = SPNG_FMT_RGB8
        rgb_format = "RGB"
        flags = SPNG_DECODE_TRNS
    elif ihdr.color_type==SPNG_COLOR_TYPE_GRAYSCALE_ALPHA:
        fmt = SPNG_FMT_RGBA8
        rgb_format = "RGBA"
        flags = SPNG_DECODE_TRNS
    else:
        raise ValueError("cannot handle color type %s" % COLOR_TYPE_STR.get(ihdr.color_type, ihdr.color_type))

    if check_error(spng_decoded_image_size(ctx, fmt, &out_size),
                   "failed to get decoded image size"):
        close()
        return None
    if out_size>MAX_SIZE:
        log.error("Error: spng image size %i is too big", out_size)
        log.error(" maximum size supported is %i", MAX_SIZE)
        close()
        return None

    cdef MemBuf membuf = getbuf(out_size, 0)
    cdef uintptr_t ptr = <uintptr_t> membuf.get_mem()
    with nogil:
        r = spng_decode_image(ctx, <void *> ptr, out_size, fmt, flags)
    if check_error(r, "failed to decode image"):
        close()
        return None

    close()
    return memoryview(membuf), rgb_format, ihdr.width, ihdr.height


def selftest(full=False) -> None:
    log("spng version %s selftest" % (get_version(),))
    from xpra.codecs.checks import TEST_PICTURES   # pylint: disable=import-outside-toplevel
    for size, samples in TEST_PICTURES["png"].items():
        for cdata, options in samples:
            assert decompress(cdata, typedict(options)), f"failed to decompress {size} png"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Iterable
from collections.abc import Sequence

from xpra.util.str_fn import csv
from xpra.log import Logger

log = Logger("audio")

VORBIS = "vorbis"
AAC = "aac"
FLAC = "flac"
MP3 = "mp3"
WAV = "wav"
OPUS = "opus"
WAVPACK = "wavpack"

OGG = "ogg"
MKA = "mka"
MPEG4 = "mpeg4"
ID3V2 = "id3v2"
# RTP = "rtp"
RAW = "raw"

# stream compression
LZ4 = "lz4"
# LZO = "lzo"

FLAC_OGG = FLAC + "+" + OGG
OPUS_OGG = OPUS + "+" + OGG
VORBIS_OGG = VORBIS + "+" + OGG
OPUS_MKA = OPUS + "+" + MKA
# OPUS_RTP    = OPUS+"+"+RTP
VORBIS_MKA = VORBIS + "+" + MKA
AAC_MPEG4 = AAC + "+" + MPEG4
WAV_LZ4 = WAV + "+" + LZ4
# WAV_LZO     = WAV+"+"+LZO
MP3_MPEG4 = MP3 + "+" + MPEG4
MP3_ID3V2 = MP3 + "+" + ID3V2


# used for parsing codec names specified on the command line:
def audio_option_or_all(name: str, options, all_values: Iterable[str]) -> Sequence[str]:
    log("audio_option_or_all%s", (name, options, all_values))
    if not options:
        v = list(all_values)  # not specified on command line: use default
    else:
        v = []
        invalid_options = []
        for x in options:
            # options is a list, but it may have csv embedded:
            for o in x.split(","):
                o = o.strip()
                if o not in all_values:
                    invalid_options.append(o)
                else:
                    v.append(o)
        if invalid_options:
            if all_values:
                log.warn(f"Warning: invalid values for {name!r}: " + csv(invalid_options))
                log.warn(" valid options: " + csv(all_values))
            else:
                log.warn(f"Warning: no {name!r} available")
    log("%s=%s", name, csv(v))
    return tuple(v)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import Any

from xpra.os_util import gi_import
from xpra.exit_codes import ExitCode
from xpra.platform.features import REINIT_WINDOWS
from xpra.platform.gui import (
    get_vrefresh,
    get_antialias_info, get_icc_info, get_display_icc_info, show_desktop, get_cursor_size,
    get_xdpi, get_ydpi, get_number_of_desktops, get_desktop_names, get_wm_name,
)
from xpra.scripts.main import check_display
from xpra.scripts.config import FALSE_OPTIONS
from xpra.net.common import MAX_PACKET_SIZE, PacketType
from xpra.common import (
    adjust_monitor_refresh_rate, get_refresh_rate_for_value,
    FULL_INFO, SYNC_ICC, NotificationID,
)
from xpra.util.parsing import (
    parse_scaling, scaleup_value, scaledown_value, fequ, r4cmp,
    MIN_SCALING, MAX_SCALING, SCALING_EMBARGO_TIME,
)
from xpra.util.objects import typedict
from xpra.util.screen import log_screen_sizes
from xpra.util.env import envint
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("screen")
workspacelog = Logger("client", "workspace")
scalinglog = Logger("scaling")

MONITOR_CHANGE_REINIT = envint("XPRA_MONITOR_CHANGE_REINIT")


class DisplayClient(StubClientMixin):
    """
    Utility mixin for clients that handle a desktop / display
    Adds client-side scaling handling
    """
    __signals__ = ["scaling-changed"]
    PREFIX = "display"

    def __init__(self):
        check_display()
        self.dpi = 0
        self.can_scale = False
        self.initial_scaling = 1, 1
        self.xscale, self.yscale = self.initial_scaling
        self.scale_change_embargo = float("inf")
        self.refresh_rate = ""
        self.desktop_fullscreen = False
        self.desktop_scaling = False
        self.screen_size_change_timer = 0

        self.server_desktop_size = None
        self.server_actual_desktop_size = None
        self.server_max_desktop_size = None
        self.server_display = None
        self.server_randr = False
        self.server_opengl = None
        self.server_multi_monitors = False
        self.server_monitors = {}
        self.log_screen_info = True

    def init(self, opts) -> None:
        self.desktop_fullscreen = opts.desktop_fullscreen
        self.desktop_scaling = opts.desktop_scaling
        self.refresh_rate = opts.refresh_rate
        self.dpi = int(opts.dpi)
        self.can_scale = opts.desktop_scaling not in FALSE_OPTIONS
        scalinglog("can_scale(%s)=%s", opts.desktop_scaling, self.can_scale)
        if self.can_scale:
            self.parse_scaling(opts.desktop_scaling)

    def parse_scaling(self, desktop_scaling: str) -> None:
        root_w, root_h = self.get_root_size()
        self.initial_scaling = parse_scaling(desktop_scaling, root_w, root_h, MIN_SCALING, MAX_SCALING)
        self.xscale, self.yscale = self.initial_scaling
        scalinglog("scaling(%s)=%s", self.initial_scaling, (self.xscale, self.yscale))

    def cleanup(self) -> None:
        ssct = self.screen_size_change_timer
        if ssct:
            self.screen_size_change_timer = 0
            GLib.source_remove(ssct)

    def get_screen_sizes(self, xscale=1, yscale=1):
        raise NotImplementedError()

    def get_root_size(self):
        raise NotImplementedError()

    def get_info(self) -> dict[str, Any]:
        screen = self.get_screen_caps()
        screen["scaling"] = self.get_scaling_caps()
        screen["dpi"] = self.get_dpi_caps()
        return {
            "screen": screen,
        }

    ######################################################################
    # hello:
    def get_caps(self) -> dict[str, Any]:
        caps: dict[str, Any] = {
            "show-desktop": True,
            "vrefresh": self.get_vrefresh(),
        }
        if FULL_INFO > 0:
            wm_name = get_wm_name()
            if wm_name:
                caps["wm_name"] = wm_name

        self._last_screen_settings = self.get_screen_settings()
        root_w, root_h, sss, ndesktops, desktop_names, u_root_w, u_root_h = self._last_screen_settings[:7]
        if u_root_w and u_root_h:
            caps["desktop_size"] = self.cp(u_root_w, u_root_h)
        if ndesktops:
            caps["desktops"] = ndesktops
            caps["desktop.names"] = tuple(desktop_names)

        ss = self.get_screen_sizes()
        self._current_screen_sizes = ss

        if self.log_screen_info:
            log.info(" desktop size is %sx%s:", u_root_w, u_root_h)
            log_screen_sizes(u_root_w, u_root_h, ss)
        if self.xscale != 1 or self.yscale != 1:
            caps["screen_sizes.unscaled"] = ss
            caps["desktop_size.unscaled"] = u_root_w, u_root_h
            if self.log_screen_info:
                root_w, root_h = self.cp(u_root_w, u_root_h)
                if fequ(self.xscale, self.yscale):
                    sinfo = "%i%%" % round(self.xscale * 100)
                else:
                    sinfo = "%i%% x %i%%" % (round(self.xscale * 100), round(self.yscale * 100))
                scaled_up = u_root_w > root_w or u_root_h > root_h
                log.info(" %sscaled to %s, virtual screen size: %ix%i",
                         "up" if scaled_up else "down", sinfo, root_w, root_h)
                log_screen_sizes(root_w, root_h, sss)
        else:
            sss = ss
        caps["screen_sizes"] = sss
        monitors = self.get_monitors_info()
        caps["monitors"] = adjust_monitor_refresh_rate(self.refresh_rate, monitors)
        caps.update(self.get_screen_caps())
        caps["dpi"] = self.get_dpi_caps()
        caps["screen-scaling"] = self.get_scaling_caps()
        if self.desktop_fullscreen:
            caps["desktop-fullscreen"] = True
        self.log_screen_info = False        # don't log it again
        return caps

    def get_dpi_caps(self) -> dict[str, Any]:
        # command line (or config file) override supplied:
        caps = {}
        dpi = 0
        if self.dpi > 0:
            # scale it:
            dpi = round((self.cx(self.dpi) + self.cy(self.dpi)) / 2.0)
        else:
            # not supplied, use platform detection code:
            # platforms may also provide per-axis dpi (later win32 versions do)
            xdpi = self.get_xdpi()
            ydpi = self.get_ydpi()
            log("xdpi=%i, ydpi=%i", xdpi, ydpi)
            if xdpi > 0 and ydpi > 0:
                xdpi = self.cx(xdpi)
                ydpi = self.cy(ydpi)
                dpi = round((xdpi + ydpi) / 2.0)
                caps = {
                    "x": xdpi,
                    "y": ydpi,
                }
        if dpi:
            caps[""] = dpi
        log("get_dpi_caps()=%s", caps)
        return caps

    def get_scaling_caps(self) -> dict[str, Any]:
        return {
            "": True,
            "enabled": self.xscale != 1 or self.yscale != 1,
            "values": (int(1000 * self.xscale), int(1000 * self.yscale)),
        }

    def get_vrefresh(self) -> int:
        return get_vrefresh()

    def get_screen_caps(self) -> dict[str, Any]:
        caps = {
            "vrefresh": self.get_vrefresh(),
            "cursor": {
                "size": int(2 * get_cursor_size() / (self.xscale + self.yscale)),
            },
        }
        aa = get_antialias_info()
        if aa:
            caps["antialias"] = aa
        return caps

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.server_display = c.strget("display")
        self.server_desktop_size = c.intpair("desktop_size")
        log("server desktop size=%s", self.server_desktop_size)
        self.server_max_desktop_size = c.intpair("max_desktop_size", (2 ** 15, 2 ** 15))
        self.server_actual_desktop_size = c.intpair("actual_desktop_size")
        log("server actual desktop size=%s", self.server_actual_desktop_size)
        self.server_randr = c.boolget("resize_screen")
        log("server has randr: %s", self.server_randr)
        self.server_opengl = c.dictget("opengl")
        Logger("screen", "opengl")("server opengl=%s", self.server_opengl)
        self.server_multi_monitors = c.boolget("multi-monitors", False)
        self.server_monitors = c.dictget("monitors")
        log("server multi-monitors=%s, monitors=%s", self.server_multi_monitors, self.server_monitors)
        return True

    def process_ui_capabilities(self, c: typedict) -> None:
        self.server_is_desktop = c.boolget("shadow") or c.boolget("desktop") or c.boolget("monitor")
        log("process_ui_capabilities(%s) desktop=%s, monitor=%s", c, c.boolget("desktop"), c.boolget("monitor"))
        if not c.boolget("monitor"):
            skip_vfb_size_check = False  # if we decide not to use scaling, skip warnings
            if not fequ(self.xscale, 1.0) or not fequ(self.yscale, 1.0):
                # scaling is used, make sure that we need it and that the server can support it
                # (without rounding support, size-hints can cause resize loops)
                if self.server_is_desktop and not self.desktop_fullscreen:
                    # don't honour auto mode in this case
                    if self.desktop_scaling == "auto":
                        log.info(" not scaling a %s server", c.strget("type", "shadow"))
                        skip_vfb_size_check = self.xscale > 1 or self.yscale > 1
                        self.scale_change_embargo = 0
                        self.scalingoff()
            if self.can_scale:
                self.may_adjust_scaling()
            if not self.server_is_desktop and not skip_vfb_size_check and self.server_max_desktop_size:
                avail_w, avail_h = self.server_max_desktop_size
                root_w, root_h = self.get_root_size()
                log("validating server_max_desktop_size=%s vs root size=%s",
                    self.server_max_desktop_size, (root_w, root_h))
                if self.cx(root_w) != root_w or self.cy(root_h) != root_h:
                    log(" root size scaled to %s", (self.cx(root_w), self.cy(root_h)))
                if self.cx(root_w) > (avail_w + 1) or self.cy(root_h) > (avail_h + 1):
                    log.warn("Server's virtual screen is too small")
                    log.warn(" server: %sx%s vs client: %sx%s", avail_w, avail_h, self.cx(root_w), self.cy(root_h))
                    log.warn(" you may see strange behavior,")
                    log.warn(" please see https://github.com/Xpra-org/xpra/blob/master/docs/Usage/Xdummy.md")
        # now that we have the server's screen info, allow scale changes:
        self.scale_change_embargo = 0
        self.set_max_packet_size()
        self.send_icc_data()

    def send_icc_data(self) -> None:
        if SYNC_ICC and "configure-display" in self.server_packet_types:
            # it is now safe to send the colourspace data if we have any:
            icc = self.get_icc_info()
            dicc = self.get_display_icc_info()
            if icc or dicc:
                self.send("configure-display", {
                    "icc": {
                        "global": icc,
                        "display": dicc,
                    },
                })

    # noinspection PyUnreachableCode
    def set_max_packet_size(self) -> None:
        p = self._protocol
        if not p or p.TYPE != "xpra":
            return
        root_w, root_h = self.cp(*self.get_root_size())
        maxw, maxh = root_w, root_h
        try:
            server_w, server_h = self.server_actual_desktop_size
            maxw = max(root_w, server_w)
            maxh = max(root_h, server_h)
        except (TypeError, ValueError):
            pass
        if maxw <= 0 or maxh <= 0 or maxw >= 32768 or maxh >= 32768:
            message = "invalid maximum desktop size: %ix%i" % (maxw, maxh)
            log(message)
            self.quit(ExitCode.INTERNAL_ERROR)
            return
        if maxw >= 16384 or maxh >= 16384:
            log.warn("Warning: the desktop size is extremely large: %ix%i", maxw, maxh)
        # max packet size to accommodate
        # a full screen RGBX (32 bits) uncompressed image
        # also with enough headroom for some metadata (4k)
        # we can't assume to have a real ClientConnection object:
        p.max_packet_size = max(MAX_PACKET_SIZE, maxw * maxh * 4 + 4 * 1024)
        p.abs_max_packet_size = maxw * maxh * 4 * 4 + 4 * 1024
        log("maximum packet size set to %i", p.max_packet_size)

    def has_transparency(self) -> bool:
        return False

    def get_icc_info(self) -> dict[str, Any]:
        return get_icc_info()

    def get_display_icc_info(self) -> dict[str, Any]:
        return get_display_icc_info()

    def get_monitors_info(self):
        return {}

    def _process_show_desktop(self, packet: PacketType) -> None:
        show = packet[1]
        log("calling %s(%s)", show_desktop, show)
        show_desktop(show)

    def _process_desktop_size(self, packet: PacketType) -> None:
        root_w, root_h, max_w, max_h = packet[1:5]
        log("server has resized the desktop to: %sx%s (max %sx%s)", root_w, root_h, max_w, max_h)
        self.server_max_desktop_size = max_w, max_h
        self.server_actual_desktop_size = root_w, root_h
        if self.can_scale:
            self.may_adjust_scaling()

    def may_adjust_scaling(self) -> None:
        log("may_adjust_scaling() server_is_desktop=%s, desktop_fullscreen=%s",
            self.server_is_desktop, self.desktop_fullscreen)
        if self.server_is_desktop and not self.desktop_fullscreen:
            # don't try to make it fit
            return
        assert self.can_scale
        max_w, max_h = self.server_max_desktop_size  # ie: server limited to 8192x4096?
        w, h = self.get_root_size()  # ie: 5760, 2160
        sw, sh = self.cp(w, h)  # ie: upscaled to: 11520x4320
        scalinglog("may_adjust_scaling() server max desktop size=%s, server actual desktop size=%s",
                   self.server_max_desktop_size, self.server_actual_desktop_size)
        scalinglog("may_adjust_scaling() client root size=%s", self.get_root_size())
        scalinglog(" scaled client root size using %sx%s: %s", self.xscale, self.yscale, (sw, sh))

        # server size is too small for the client screen size with the current scaling value,
        # calculate the minimum scaling to fit it:

        def clamp(v) -> int | float:
            return max(MIN_SCALING, min(MAX_SCALING, v))

        if self.desktop_fullscreen:
            sw, sh = self.server_actual_desktop_size
            x = clamp(w / sw)
            y = clamp(h / sh)
        else:
            if sw < (max_w + 1) and sh < (max_h + 1):
                # no change needed
                return
            x = clamp(w / max_w)
            y = clamp(h / max_h)
            # avoid wonky scaling:
            if not 0.75 < x / y < 1.25:
                x = y = min(x, y)

        def mint(v) -> int | float:
            # prefer `int` to `float`,
            # and even tolerate a 0.1% difference to get it:
            if round(v) * 1000 == round(v * 1000):
                return int(v)
            return v

        self.xscale = mint(x)
        self.yscale = mint(y)
        scalinglog(" xscale=%s, yscale=%s", self.xscale, self.yscale)
        # to use the same scale for both axes:
        # self.xscale = mint(max(x, y))
        # self.yscale = self.xscale
        summary = "Desktop scaling adjusted to accommodate the server"
        xstr = ("%.3f" % self.xscale).rstrip("0")
        ystr = ("%.3f" % self.yscale).rstrip("0")
        messages = [
            "server desktop size is %ix%i" % (max_w, max_h),
            "using scaling factor %s x %s" % (xstr, ystr),
        ]
        self.may_notify(NotificationID.SCALING, summary, "\n".join(messages), icon_name="scaling")
        scalinglog.warn("Warning: %s", summary)
        for m in messages:
            scalinglog.warn(" %s", m)
        self.emit("scaling-changed")

    ######################################################################
    # screen scaling:
    def fsx(self, v) -> float:
        """ convert X coordinate from server to client """
        return v * self.xscale

    def fsy(self, v) -> float:
        """ convert Y coordinate from server to client """
        return v * self.yscale

    def sx(self, v) -> int:
        """ convert X coordinate from server to client """
        return round(self.fsx(v))

    def sy(self, v) -> int:
        """ convert Y coordinate from server to client """
        return round(self.fsy(v))

    def srect(self, x, y, w, h) -> tuple[int, int, int, int]:
        """ convert rectangle coordinates from server to client """
        return self.sx(x), self.sy(y), self.sx(w), self.sy(h)

    def sp(self, x, y) -> tuple[int, int]:
        """ convert X,Y coordinates from server to client """
        return self.sx(x), self.sy(y)

    def cx(self, v) -> int:
        """ convert X coordinate from client to server """
        return round(v / self.xscale)

    def cy(self, v) -> int:
        """ convert Y coordinate from client to server """
        return round(v / self.yscale)

    def crect(self, x, y, w, h) -> tuple[int, int, int, int]:
        """ convert rectangle coordinates from client to server """
        return self.cx(x), self.cy(y), self.cx(w), self.cy(h)

    def cp(self, x, y) -> tuple[int, int]:
        """ convert X,Y coordinates from client to server """
        return self.cx(x), self.cy(y)

    ######################################################################
    # desktop, screen and scaling:
    def desktops_changed(self, *args) -> None:
        workspacelog("desktops_changed%s", args)
        self.screen_size_changed(*args)

    def workspace_changed(self, *args) -> None:
        workspacelog("workspace_changed%s", args)
        for win in self._id_to_window.values():
            win.workspace_changed()

    def screen_size_changed(self, *args) -> None:
        log("screen_size_changed(%s) timer=%s", args, self.screen_size_change_timer)
        if self.screen_size_change_timer:
            return
        # update via timer so the data is more likely to be final (up to date) when we query it,
        # some properties (like _NET_WORKAREA for X11 clients via posix gui "ClientExtras") may
        # trigger multiple calls to screen_size_changed, delayed by some amount
        # (sometimes up to 1s..)
        delay = 1000
        # if we are suspending, wait longer:
        # (better chance that the suspend-resume cycle will have completed)
        if self._suspended_at > 0 and self._suspended_at - monotonic() < 5 * 1000:
            delay = 5 * 1000
        self.screen_size_change_timer = GLib.timeout_add(delay, self.do_process_screen_size_change)

    def do_process_screen_size_change(self) -> None:
        self.screen_size_change_timer = 0
        self.update_screen_size()
        log("do_process_screen_size_change() MONITOR_CHANGE_REINIT=%s, REINIT_WINDOWS=%s",
            MONITOR_CHANGE_REINIT, REINIT_WINDOWS)
        if MONITOR_CHANGE_REINIT and MONITOR_CHANGE_REINIT == "0":
            return
        if MONITOR_CHANGE_REINIT or REINIT_WINDOWS:
            log.info("screen size change: will reinit the windows")
            self.reinit_windows()
            self.reinit_window_icons()

    def get_screen_settings(self) -> tuple:
        u_root_w, u_root_h = self.get_root_size()
        root_w, root_h = self.cp(u_root_w, u_root_h)
        self._current_screen_sizes = self.get_screen_sizes()
        sss = self.get_screen_sizes(self.xscale, self.yscale)
        ndesktops = get_number_of_desktops()
        desktop_names = get_desktop_names()
        log("get_screen_settings() sizes=%s, %s desktops: %s", sss, ndesktops, desktop_names)
        if self.dpi > 0:
            # use command line value supplied, but scale it:
            xdpi = ydpi = self.dpi
            log("get_screen_settings() dpi=%s", self.dpi)
        else:
            # not supplied, use platform detection code:
            xdpi = self.get_xdpi()
            ydpi = self.get_ydpi()
            log("get_screen_settings() xdpi=%s, ydpi=%s", get_xdpi(), get_ydpi())
        xdpi = self.cx(xdpi)
        ydpi = self.cy(ydpi)
        log("get_screen_settings() scaled: xdpi=%s, ydpi=%s", xdpi, ydpi)
        vrefresh = self.get_vrefresh()
        rrate = -1
        if vrefresh > 0:
            rrate = (get_refresh_rate_for_value(self.refresh_rate, vrefresh) or 0) // 1000
        log("get_screen_settings() vrefresh=%s (from %s)", rrate, vrefresh)
        monitors = self.get_monitors_info()
        return root_w, root_h, sss, ndesktops, desktop_names, u_root_w, u_root_h, xdpi, ydpi, rrate, monitors

    def update_screen_size(self) -> None:
        self.screen_size_change_timer = 0
        screen_settings = self.get_screen_settings()
        log("update_screen_size()     new settings=%s", screen_settings)
        log("update_screen_size() current settings=%s", self._last_screen_settings)
        if self._last_screen_settings == screen_settings:
            log("screen size unchanged")
            return
        root_w, root_h, sss = screen_settings[:3]
        log.info("sending updated screen size to server: %sx%s", root_w, root_h)
        log_screen_sizes(root_w, root_h, sss)
        if "configure-display" in self.server_packet_types:
            root_w, root_h = screen_settings[:2]
            ndesktops, desktop_names = screen_settings[3:5]
            u_root_w, u_root_h = screen_settings[5:7]
            xdpi, ydpi = screen_settings[7:9]
            rrate, monitors = screen_settings[9:11]
            attrs = {
                "desktop-size": (root_w, root_h),
                "desktop-size-unscaled": (u_root_w, u_root_h),
                "monitors": monitors,
                "dpi": {"x": xdpi, "y": ydpi},
            }
            icc = self.get_icc_info()
            dicc = self.get_display_icc_info()
            if icc or dicc:
                attrs["icc"] = {
                    "global": icc,
                    "display": dicc,
                }
            if ndesktops:
                attrs["desktops"] = ndesktops
                attrs["desktop-names"] = desktop_names or ()
            if rrate:
                attrs["vrefresh"] = rrate
            log(f"configure-display: {attrs}")
            self.send("configure-display", attrs)
        self._last_screen_settings = screen_settings
        # update the max packet size (may have gone up):
        self.set_max_packet_size()

    def get_xdpi(self) -> int:
        return get_xdpi()

    def get_ydpi(self) -> int:
        return get_ydpi()

    def scaleup(self) -> None:
        scaling = max(self.xscale, self.yscale)
        options = scaleup_value(scaling)
        scalinglog("scaleup() options=%s", options)
        if options:
            self._scaleto(min(options))

    def scaledown(self) -> None:
        scaling = min(self.xscale, self.yscale)
        options = scaledown_value(scaling)
        scalinglog("scaledown() options=%s", options)
        if options:
            self._scaleto(max(options))

    def _scaleto(self, new_scaling) -> None:
        scaling = max(self.xscale, self.yscale)
        scalinglog("_scaleto(%s) current value=%s", r4cmp(new_scaling, 1000) / 1000.0, r4cmp(scaling, 1000) / 1000.0)
        if new_scaling > 0:
            self.scale_change(new_scaling / scaling, new_scaling / scaling)

    def scalingoff(self) -> None:
        self.scaleset(1, 1)

    def scalereset(self) -> None:
        self.scaleset(*self.initial_scaling)

    def scaleset(self, xscale=1.0, yscale=1.0) -> None:
        scalinglog("scaleset(%s, %s) current scaling: %s, %s", xscale, yscale, self.xscale, self.yscale)
        self.scale_change(xscale / self.xscale, yscale / self.yscale)

    def scale_change(self, xchange=1.0, ychange=1.0) -> None:
        scalinglog("scale_change(%s, %s)", xchange, ychange)
        if self.server_is_desktop and self.desktop_fullscreen:
            scalinglog("scale_change(%s, %s) ignored, fullscreen shadow mode is active", xchange, ychange)
            return
        if not self.can_scale:
            scalinglog("scale_change(%s, %s) ignored, scaling is disabled", xchange, ychange)
            return
        if self.screen_size_change_timer:
            scalinglog("scale_change(%s, %s) screen size change is already pending", xchange, ychange)
            return
        if monotonic() < self.scale_change_embargo:
            scalinglog("scale_change(%s, %s) screen size change not permitted during embargo time - try again",
                       xchange, ychange)
            return

        def clamp(v) -> int | float:
            return max(MIN_SCALING, min(MAX_SCALING, v))

        xscale = clamp(self.xscale * xchange)
        yscale = clamp(self.yscale * ychange)
        scalinglog("scale_change xscale: clamp(%s*%s)=%s", self.xscale, xchange, xscale)
        scalinglog("scale_change yscale: clamp(%s*%s)=%s", self.yscale, ychange, yscale)
        if fequ(xscale, self.xscale) and fequ(yscale, self.yscale):
            scalinglog("scaling unchanged: %sx%s", self.xscale, self.yscale)
            return
        # re-calculate change values against clamped scale:
        xchange = xscale / self.xscale
        ychange = yscale / self.yscale
        # check against maximum server supported size:
        maxw, maxh = self.server_max_desktop_size
        root_w, root_h = self.get_root_size()
        sw = int(root_w / xscale)
        sh = int(root_h / yscale)
        scalinglog("scale_change root size=%s x %s, scaled to %s x %s", root_w, root_h, sw, sh)
        scalinglog("scale_change max server desktop size=%s x %s", maxw, maxh)
        if not self.server_is_desktop and (sw > (maxw + 1) or sh > (maxh + 1)):
            # would overflow..
            summary = "Invalid Scale Factor"
            messages = [
                "cannot scale by %i%% x %i%% or lower" % (round(100 * xscale), round(100 * yscale)),
                "the scaled client screen %i x %i -> %i x %i" % (root_w, root_h, sw, sh),
                " would overflow the server's screen: %i x %i" % (maxw, maxh),
            ]
            self.may_notify(NotificationID.SCALING, summary, "\n".join(messages), icon_name="scaling")
            scalinglog.warn("Warning: %s", summary)
            for m in messages:
                scalinglog.warn(" %s", m)
            return
        self.xscale = xscale
        self.yscale = yscale
        scalinglog("scale_change new scaling: %sx%s, change: %sx%s", self.xscale, self.yscale, xchange, ychange)
        self.scale_reinit(xchange, ychange)

    def scale_reinit(self, xchange=1.0, ychange=1.0) -> None:
        # wait at least one second before changing again:
        self.scale_change_embargo = monotonic() + SCALING_EMBARGO_TIME
        if fequ(self.xscale, self.yscale):
            scalinglog.info("setting scaling to %i%%:", round(100 * self.xscale))
        else:
            scalinglog.info("setting scaling to %i%% x %i%%:", round(100 * self.xscale), round(100 * self.yscale))
        self.update_screen_size()
        # re-initialize all the windows with their new size
        for win in self._id_to_window.values():
            if hasattr(win, "_xscale"):
                win._xscale *= xchange
            if hasattr(win, "_yscale"):
                win._yscale *= ychange

        def new_size_fn(w, h) -> tuple[int, int]:
            minx, miny = 16384, 16384
            if self.max_window_size != (0, 0):
                minx, miny = self.max_window_size
            return max(1, min(minx, round(w * xchange))), max(1, min(miny, round(h * ychange)))

        self.resize_windows(new_size_fn)
        self.reinit_window_icons()
        self.emit("scaling-changed")

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets("show-desktop", "desktop_size", main_thread=True)



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import signal
from collections.abc import Callable

from xpra.os_util import POSIX, gi_import
from xpra.util.system import SIGNAMES
from xpra.util.io import stderr_print, get_util_logger

_glib_unix_signals: dict[int, int] = {}


def quit_on_signals(commandtype: str = "") -> None:
    Gtk = gi_import("Gtk")

    def signal_handler(_signum: int) -> None:
        Gtk.main_quit()

    register_os_signals(signal_handler, commandtype)


def register_os_signals(callback: Callable[[int], None],
                        commandtype: str = "",
                        signals=(signal.SIGINT, signal.SIGTERM)) -> None:
    for signum in signals:
        register_os_signal(callback, commandtype, signum)


def register_os_signal(callback: Callable[[int], None],
                       commandtype: str = "",
                       signum: signal.Signals = signal.SIGINT) -> None:
    GLib = gi_import("GLib")
    signame = SIGNAMES.get(signum, str(signum))
    i_signum = int(signum)

    def write_signal() -> None:
        if not commandtype:
            return
        try:
            stderr_print()
            cstr = ""
            if commandtype:
                cstr = commandtype + " "
            get_util_logger().info(f"{cstr}got signal {signame}")
        except OSError:
            pass

    def do_handle_signal() -> None:
        callback(i_signum)

    if POSIX:
        # replace the previous definition if we had one:
        current = _glib_unix_signals.get(signum, None)
        if current:
            GLib.source_remove(current)

        def handle_signal(_signum) -> bool:
            write_signal()
            GLib.idle_add(do_handle_signal)
            return True

        source_id = GLib.unix_signal_add(GLib.PRIORITY_HIGH, i_signum, handle_signal, signum)
        _glib_unix_signals[signum] = source_id
    else:
        def os_signal(_signum, _frame) -> None:
            write_signal()
            GLib.idle_add(do_handle_signal)

        signal.signal(signum, os_signal)


def register_SIGUSR_signals(commandtype: str = "Server") -> None:
    if not POSIX:
        return
    from xpra.util.pysystem import dump_all_frames, dump_gc_frames
    log = get_util_logger()

    def sigusr1(_sig: int) -> None:
        log.info("SIGUSR1")
        dump_all_frames(log.info)

    def sigusr2(_sig: int) -> None:
        log.info("SIGUSR2")
        dump_gc_frames(log.info)

    register_os_signals(sigusr1, commandtype, (signal.SIGUSR1,))
    register_os_signals(sigusr2, commandtype, (signal.SIGUSR2,))


def install_signal_handlers(sstr: str, signal_handler: Callable[[int], None]) -> None:
    # only register the glib signal handler
    # once the main loop is running,
    # before that we just trigger a KeyboardInterrupt
    def do_install_signal_handlers() -> None:
        register_os_signals(signal_handler, sstr)
        register_SIGUSR_signals(sstr)

    GLib = gi_import("GLib")
    GLib.idle_add(do_install_signal_handlers)
